import json
from shlex import split
from textwrap import dedent
from unittest import TestCase

from lxml import etree

from pcs.common.interface.dto import to_dict
from pcs.common.pacemaker.defaults import CibDefaultsDto
from pcs.common.pacemaker.nvset import CibNvpairDto, CibNvsetDto
from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.str_tools import format_list, format_plural
from pcs.common.types import CibRuleExpressionType, CibRuleInEffectStatus

from pcs_test.tools import metadata_dto
from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.bin_mock import get_mock_settings
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    is_minimum_cib_schema_version,
    skip_unless_crm_rule,
    skip_unless_pacemaker_supports_rsc_and_op_rules,
    write_data_to_tmpfile,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import XmlManipulation

empty_cib = rc("cib-empty.xml")
empty_cib_rules = rc("cib-empty-3.9.xml")
RULE_ARGV_DEPRECATED = (
    "Deprecation Warning: Specifying a rule as multiple arguments is "
    "deprecated and might be removed in a future release, specify the rule as "
    "a single string instead\n"
)
DEFAULTS_MAY_BE_OVERRIDDEN = (
    "Warning: Defaults do not apply to resources which override them "
    "with their own defined values\n"
)
CIB_HAS_BEEN_UPGRADED = "CIB has been upgraded to the latest schema version.\n"
META_NOT_VALIDATED = (
    "Warning: Meta attributes of bundle / clone / group / resource are not "
    "validated\n"
)


def fixture_warning_unknown_meta_primitive(unknown_meta):
    known_meta = format_list(
        sorted(
            set(
                metadata_dto.FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META
                + metadata_dto.FIXTURE_KNOWN_META_NAMES_STONITH_META
            )
        )
    )
    attributes = format_plural(unknown_meta, "attribute")
    have = format_plural(unknown_meta, "has")
    return (
        f"Warning: Resource / stonith meta {attributes} "
        f"{format_list(unknown_meta)} {have} no effect on cluster resource "
        f"handling, meta attributes with effect: {known_meta}\n"
    )


def fixture_defaults_dto(prefix, include_expired):
    meta_attributes = [
        CibNvsetDto(
            id=f"{prefix}-set1",
            options={},
            rule=CibRuleExpressionDto(
                id=f"{prefix}-set1-rule",
                type=CibRuleExpressionType.RULE,
                in_effect=CibRuleInEffectStatus.NOT_YET_IN_EFFECT,
                options={"boolean-op": "and"},
                date_spec=None,
                duration=None,
                expressions=[
                    CibRuleExpressionDto(
                        id=f"{prefix}-set1-rule-expr",
                        type=CibRuleExpressionType.DATE_EXPRESSION,
                        in_effect=CibRuleInEffectStatus.UNKNOWN,
                        options={"operation": "gt", "start": "3000-01-01"},
                        date_spec=None,
                        duration=None,
                        expressions=[],
                        as_string="date gt 3000-01-01",
                    )
                ],
                as_string="date gt 3000-01-01",
            ),
            nvpairs=[
                CibNvpairDto(
                    id=f"{prefix}-set1-name1", name="name1", value="value1"
                )
            ],
        ),
        CibNvsetDto(
            id=f"{prefix}-set2",
            options={},
            rule=CibRuleExpressionDto(
                id=f"{prefix}-set2-rule",
                type="RULE",
                in_effect=CibRuleInEffectStatus.EXPIRED,
                options={"boolean-op": "and"},
                date_spec=None,
                duration=None,
                expressions=[
                    CibRuleExpressionDto(
                        id=f"{prefix}-set2-rule-expr",
                        type=CibRuleExpressionType.DATE_EXPRESSION,
                        in_effect=CibRuleInEffectStatus.UNKNOWN,
                        options={"end": "1000-01-01", "operation": "lt"},
                        date_spec=None,
                        duration=None,
                        expressions=[],
                        as_string="date lt 1000-01-01",
                    )
                ],
                as_string="date lt 1000-01-01",
            ),
            nvpairs=[
                CibNvpairDto(
                    id=f"{prefix}-set2-name2", name="name2", value="value2"
                )
            ],
        ),
        CibNvsetDto(
            id=f"{prefix}-set3",
            nvpairs=[
                CibNvpairDto(
                    id=f"{prefix}-set3-name3", name="name3", value="value3"
                )
            ],
            options={},
            rule=CibRuleExpressionDto(
                id=f"{prefix}-set3-rule",
                type="RULE",
                in_effect=CibRuleInEffectStatus.IN_EFFECT,
                options={"boolean-op": "and"},
                date_spec=None,
                duration=None,
                expressions=[
                    CibRuleExpressionDto(
                        id=f"{prefix}-set3-rule-expr",
                        type=CibRuleExpressionType.DATE_EXPRESSION,
                        in_effect=CibRuleInEffectStatus.UNKNOWN,
                        options={
                            "end": "3000-01-01",
                            "operation": "in_range",
                            "start": "1000-01-01",
                        },
                        date_spec=None,
                        duration=None,
                        expressions=[],
                        as_string="date in_range 1000-01-01 to 3000-01-01",
                    )
                ],
                as_string="date in_range 1000-01-01 to 3000-01-01",
            ),
        ),
    ]
    return CibDefaultsDto(
        instance_attributes=[],
        meta_attributes=(
            meta_attributes
            if include_expired
            else [
                dto
                for dto in meta_attributes
                if not dto.rule
                or dto.rule.in_effect != CibRuleInEffectStatus.EXPIRED
            ]
        ),
    )


class TestDefaultsMixin:
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_cib_options")
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        self.pcs_runner.mock_settings = get_mock_settings("crm_resource_exec")
        self.maxDiff = None

    def tearDown(self):
        self.temp_cib.close()


class DefaultsConfigMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []
    prefix = ""

    def _prepare_cib_data(self, xml_template, cib=None):
        xml_rsc = xml_template.format(tag="rsc")
        xml_op = xml_template.format(tag="op")
        xml_manip = XmlManipulation.from_file(cib or empty_cib)
        xml_manip.append_to_first_tag_name("configuration", xml_rsc, xml_op)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

    def test_success(self):
        xml_rsc = """
            <rsc_defaults>
                <meta_attributes id="rsc-set1" score="10">
                    <nvpair id="rsc-set1-nv1" name="name1" value="rsc1"/>
                    <nvpair id="rsc-set1-nv2" name="name2" value="rsc2"/>
                </meta_attributes>
                <meta_attributes id="rsc-setA">
                    <nvpair id="rsc-setA-nv1" name="name1" value="rscA"/>
                    <nvpair id="rsc-setA-nv2" name="name2" value="rscB"/>
                </meta_attributes>
            </rsc_defaults>
        """
        xml_op = """
            <op_defaults>
                <meta_attributes id="op-set1" score="10">
                    <nvpair id="op-set1-nv1" name="name1" value="op1"/>
                    <nvpair id="op-set1-nv2" name="name2" value="op2"/>
                </meta_attributes>
                <meta_attributes id="op-setA">
                    <nvpair id="op-setA-nv1" name="name1" value="opA"/>
                    <nvpair id="op-setA-nv2" name="name2" value="opB"/>
                </meta_attributes>
            </op_defaults>
        """
        xml_manip = XmlManipulation.from_file(empty_cib)
        xml_manip.append_to_first_tag_name("configuration", xml_rsc, xml_op)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_pcs_success(
            self.cli_command,
            stdout_full=dedent(
                f"""\
                Meta Attrs: {self.prefix}-set1 score=10
                  name1={self.prefix}1
                  name2={self.prefix}2
                Meta Attrs: {self.prefix}-setA
                  name1={self.prefix}A
                  name2={self.prefix}B
            """
            ),
        )

    def test_success_rule(self):
        xml_template = """<{tag}_defaults>
            <meta_attributes id="{tag}-set1">
                <rule id="{tag}-set1-rule" boolean-op="and">
                    <rule id="{tag}-set1-rule-rule" boolean-op="or">
                        <expression id="{tag}-set1-rule-rule-expr"
                            operation="defined" attribute="attr1"
                        />
                        <expression id="{tag}-set1-rule-rule-expr-1"
                            attribute="attr2" operation="gte"
                            type="number" value="12"
                        />
                        <expression id="{tag}-set1-rule-rule-expr-2"
                            attribute="attr3" operation="lt"
                            type="version" value="3.2.1"
                        />
                        <expression id="{tag}-set1-rule-rule-expr-3"
                            attribute="attr4" operation="ne"
                            type="string" value="test"
                        />
                        <expression id="{tag}-set1-rule-rule-expr-4"
                            attribute="attr5" operation="lt" value="3"
                        />
                    </rule>
                    <rule id="{tag}-set1-rule-rule-1" boolean-op="or">
                        <date_expression id="{tag}-set1-rule-rule-1-expr"
                            operation="gt" start="2018-05-17T13:28:19"
                        />
                        <date_expression id="{tag}-set1-rule-rule-1-expr-1"
                            operation="in_range"
                            start="2019-01-01" end="2019-03-15"
                        />
                        <date_expression id="{tag}-set1-rule-rule-1-expr-2"
                             operation="in_range" start="2019-05-01"
                        >
                            <duration id="{tag}-set1-rule-rule-1-expr-2-duration"
                                months="2"
                            />
                        </date_expression>
                        <date_expression id="{tag}-set1-rule-rule-1-expr-3"
                            operation="date_spec"
                        >
                            <date_spec id="{tag}-set1-rule-rule-1-expr-3-datespec"
                                months="7-8" weekdays="6-7" years="2019"
                            />
                        </date_expression>
                        <date_expression id="{tag}-set1-rule-rule-1-expr-4"
                            operation="in_range" end="2019-12-15"
                        />
                    </rule>
                </rule>
                <nvpair id="{tag}-set1-name1" name="name1" value="value1"/>
                <nvpair id="{tag}-set1-name2" name="name2" value="value2"/>
            </meta_attributes>
        </{tag}_defaults>"""
        self._prepare_cib_data(xml_template, empty_cib_rules)

        self.assert_pcs_success(
            self.cli_command,
            stdout_full=dedent(
                f"""\
                Meta Attrs: {self.prefix}-set1
                  name1=value1
                  name2=value2
                  Rule: boolean-op=and
                    Rule: boolean-op=or
                      Expression: defined attr1
                      Expression: attr2 gte number 12
                      Expression: attr3 lt version 3.2.1
                      Expression: attr4 ne string test
                      Expression: attr5 lt 3
                    Rule: boolean-op=or
                      Expression: date gt 2018-05-17T13:28:19
                      Expression: date in_range 2019-01-01 to 2019-03-15
                      Expression: date in_range 2019-05-01 to duration
                        Duration: months=2
                      Expression:
                        Date Spec: months=7-8 weekdays=6-7 years=2019
                      Expression: date in_range to 2019-12-15
            """
            ),
        )

    xml_expired_template = """<{tag}_defaults>
        <meta_attributes id="{tag}-set1">
            <rule id="{tag}-set1-rule" boolean-op="and">
                <date_expression id="{tag}-set1-rule-expr"
                    operation="gt" start="3000-01-01"
                />
            </rule>
            <nvpair id="{tag}-set1-name1" name="name1" value="value1"/>
        </meta_attributes>
        <meta_attributes id="{tag}-set2">
            <rule id="{tag}-set2-rule" boolean-op="and">
                <date_expression id="{tag}-set2-rule-expr"
                    operation="lt" end="1000-01-01"
                />
            </rule>
            <nvpair id="{tag}-set2-name2" name="name2" value="value2"/>
        </meta_attributes>
        <meta_attributes id="{tag}-set3">
            <rule id="{tag}-set3-rule" boolean-op="and">
                <date_expression id="{tag}-set3-rule-expr"
                    operation="in_range" start="1000-01-01" end="3000-01-01"
                />
            </rule>
            <nvpair id="{tag}-set3-name3" name="name3" value="value3"/>
        </meta_attributes>
    </{tag}_defaults>"""

    @skip_unless_crm_rule()
    def test_success_rule_expired(self):
        self._prepare_cib_data(self.xml_expired_template, empty_cib_rules)
        self.assert_pcs_success(
            self.cli_command,
            stdout_full=dedent(
                f"""\
                Meta Attrs (not yet in effect): {self.prefix}-set1
                  name1=value1
                  Rule (not yet in effect): boolean-op=and
                    Expression: date gt 3000-01-01
                Meta Attrs: {self.prefix}-set3
                  name3=value3
                  Rule: boolean-op=and
                    Expression: date in_range 1000-01-01 to 3000-01-01
            """
            ),
        )

    @skip_unless_crm_rule()
    def test_success_rule_expired_all(self):
        self._prepare_cib_data(self.xml_expired_template, empty_cib_rules)
        self.assert_pcs_success(
            self.cli_command + ["--all"],
            stdout_full=dedent(
                f"""\
                Meta Attrs (not yet in effect): {self.prefix}-set1
                  name1=value1
                  Rule (not yet in effect): boolean-op=and
                    Expression: date gt 3000-01-01
                Meta Attrs (expired): {self.prefix}-set2
                  name2=value2
                  Rule (expired): boolean-op=and
                    Expression: date lt 1000-01-01
                Meta Attrs: {self.prefix}-set3
                  name3=value3
                  Rule: boolean-op=and
                    Expression: date in_range 1000-01-01 to 3000-01-01
            """
            ),
        )

    def _test_success_json(self, use_all):
        self._prepare_cib_data(self.xml_expired_template, empty_cib_rules)
        args = ["--output-format=json"]
        if use_all:
            args.append("--all")
        stdout, stderr, retval = self.pcs_runner.run(self.cli_command + args)
        self.assertEqual(
            json.loads(stdout),
            to_dict(fixture_defaults_dto(self.prefix, use_all)),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    @skip_unless_crm_rule()
    def test_success_json_format_all(self):
        self._test_success_json(use_all=True)

    @skip_unless_crm_rule()
    def test_success_json_format_not_all(self):
        self._test_success_json(use_all=False)

    def _get_as_json(self, runner, use_all):
        args = ["--output-format=json"]
        if use_all:
            args.append("--all")
        stdout, stderr, retval = runner.run(self.cli_command + args)
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        return json.loads(stdout)

    def _test_success_cmd(self, use_all):
        self._prepare_cib_data(self.xml_expired_template, empty_cib_rules)
        new_cib = get_tmp_file("tier1_cib_options_new")
        xml_manip = XmlManipulation.from_file(empty_cib_rules)
        write_data_to_tmpfile(str(xml_manip), new_cib)
        pcs_runner_new = PcsRunner(new_cib.name)
        args = ["--output-format=cmd"]
        if use_all:
            args.append("--all")
        stdout, stderr, retval = self.pcs_runner.run(self.cli_command + args)
        self.assertEqual(retval, 0)
        self.assertEqual(stderr, "")
        cmds = [
            split(cmd)[1:]
            for cmd in stdout.replace("\\\n", "").strip().split(";\n")
        ]
        for cmd in cmds:
            stdout, stderr, retval = pcs_runner_new.run(cmd)
            self.assertEqual(
                retval,
                0,
                (
                    f"Command {cmd} exited with {retval}\nstdout:\n{stdout}\n"
                    f"stderr:\n{stderr}"
                ),
            )
        self.assertEqual(
            self._get_as_json(pcs_runner_new, use_all),
            self._get_as_json(self.pcs_runner, use_all),
        )

    @skip_unless_crm_rule()
    def test_success_cmd_use_all(self):
        self._test_success_cmd(use_all=True)

    @skip_unless_crm_rule()
    def test_success_cmd_not_all(self):
        self._test_success_cmd(use_all=False)

    def _test_full_error(self, output_format):
        self.assert_pcs_fail(
            self.cli_command + [f"--output-format={output_format}", "--full"],
            stderr_full=(
                "Error: option '--full' is not compatible with "
                f"'{output_format}' output format.\n"
            ),
        )

    def test_full_json(self):
        self._test_full_error("json")

    def test_full__cmd(self):
        self._test_full_error("cmd")

    def test_full_text(self):
        self._prepare_cib_data(self.xml_expired_template, empty_cib_rules)
        self.assert_pcs_success(
            self.cli_command + ["--full"],
            stdout_full=dedent(
                f"""\
                Meta Attrs (not yet in effect): {self.prefix}-set1
                  name1=value1 (id: {self.prefix}-set1-name1)
                  Rule (not yet in effect): boolean-op=and (id: {self.prefix}-set1-rule)
                    Expression: date gt 3000-01-01 (id: {self.prefix}-set1-rule-expr)
                Meta Attrs: {self.prefix}-set3
                  name3=value3 (id: {self.prefix}-set3-name3)
                  Rule: boolean-op=and (id: {self.prefix}-set3-rule)
                    Expression: date in_range 1000-01-01 to 3000-01-01 (id: {self.prefix}-set3-rule-expr)
            """
            ),
        )


class RscDefaultsConfig(
    DefaultsConfigMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]
    prefix = "rsc"

    @skip_unless_pacemaker_supports_rsc_and_op_rules()
    def test_success_rules_rsc_op(self):
        xml = """
            <rsc_defaults>
                <meta_attributes id="X">
                    <rule id="X-rule" boolean-op="and">
                        <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                    </rule>
                    <nvpair id="X-name1" name="name1" value="value1"/>
                </meta_attributes>
            </rsc_defaults>
        """
        xml_manip = XmlManipulation.from_file(empty_cib_rules)
        xml_manip.append_to_first_tag_name("configuration", xml)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_pcs_success(
            self.cli_command,
            stdout_full=dedent(
                """\
                Meta Attrs: X
                  name1=value1
                  Rule: boolean-op=and
                    Expression: resource ::Dummy
            """
            ),
        )


class OpDefaultsConfig(
    DefaultsConfigMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]
    prefix = "op"

    @skip_unless_pacemaker_supports_rsc_and_op_rules()
    def test_success_rules_rsc_op(self):
        xml = """
            <op_defaults>
                <meta_attributes id="X">
                    <rule id="X-rule" boolean-op="and">
                        <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                        <op_expression id="X-rule-op-monitor" name="monitor"/>
                    </rule>
                    <nvpair id="X-name1" name="name1" value="value1"/>
                </meta_attributes>
            </op_defaults>
        """
        xml_manip = XmlManipulation.from_file(empty_cib_rules)
        xml_manip.append_to_first_tag_name("configuration", xml)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_pcs_success(
            self.cli_command,
            stdout_full=dedent(
                """\
                Meta Attrs: X
                  name1=value1
                  Rule: boolean-op=and
                    Expression: resource ::Dummy
                    Expression: op monitor
            """
            ),
        )


class DefaultsSetCreateMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []
    cib_tag = ""
    meta_warning_not_validated = ""

    def setUp(self):
        super().setUp()
        write_file_to_tmpfile(empty_cib, self.temp_cib)

    def test_no_args(self):
        self.assert_effect(
            self.cli_command + ["set", "create"],
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="{self.cib_tag}-meta_attributes"/>
                </{self.cib_tag}>
            """
            ),
            stderr_full=DEFAULTS_MAY_BE_OVERRIDDEN,
        )

    def test_success(self):
        self.assert_effect(
            self.cli_command
            + "set create id=mine score=10 meta name1=value1 name2=value2 --force".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="mine" score="10">
                        <nvpair id="mine-name1" name="name1" value="value1"/>
                        <nvpair id="mine-name2" name="name2" value="value2"/>
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            stderr_full=(
                self.meta_warning_not_validated + DEFAULTS_MAY_BE_OVERRIDDEN
            ),
        )

    def _assert_success_rule(self, deprecated_rule_form):
        command = (
            self.cli_command
            + "-- set create id=mine score=10 meta name1=value1 name2=value2 rule".split()
        )
        rule_str = (
            "(date gt 2018-05-17T13:28:19 or "
            "date in_range 2019-01-01 to 2019-03-15 or "
            "date in_range 2019-05-01 to duration months=2 or "
            "date-spec years=2019 months=7-8 weekdays=6-7 or "
            "date in_range to 2019-12-15)"
        )
        if deprecated_rule_form:
            full_command = command + rule_str.split()
        else:
            full_command = command + [rule_str]

        self.assert_effect(
            full_command,
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="mine" score="10">
                        <rule id="mine-rule" boolean-op="or">
                            <date_expression id="mine-rule-expr"
                                operation="gt" start="2018-05-17T13:28:19"
                            />
                            <date_expression id="mine-rule-expr-1"
                                operation="in_range"
                                start="2019-01-01" end="2019-03-15"
                            />
                            <date_expression id="mine-rule-expr-2"
                                operation="in_range" start="2019-05-01"
                            >
                                <duration id="mine-rule-expr-2-duration"
                                    months="2"
                                />
                            </date_expression>
                            <date_expression id="mine-rule-expr-3"
                                operation="date_spec"
                            >
                                <date_spec
                                    id="mine-rule-expr-3-datespec"
                                    months="7-8" weekdays="6-7" years="2019"
                                />
                            </date_expression>
                            <date_expression id="mine-rule-expr-4"
                                  operation="in_range" end="2019-12-15"
                            />
                        </rule>
                        <nvpair id="mine-name1" name="name1" value="value1"/>
                        <nvpair id="mine-name2" name="name2" value="value2"/>
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            stderr_full=(
                (RULE_ARGV_DEPRECATED if deprecated_rule_form else "")
                + CIB_HAS_BEEN_UPGRADED
                + self.meta_warning_not_validated
                + DEFAULTS_MAY_BE_OVERRIDDEN
            ),
        )

    def test_success_rule(self):
        self._assert_success_rule(False)

    def test_success_rule_deprecated_form(self):
        self._assert_success_rule(True)

    def test_rule_error_messages(self):
        self.assert_pcs_fail(
            self.cli_command
            + "set create id=mine score=10 meta name1=value1 name2=value2 rule".split()
            + [
                "(date gt 2018-05-1X or "
                "date in_range 2019-03-05 to 2019-01-11 or "
                "date in_range 2019-05-0X to duration months=2 months=3a x=y or "
                "date-spec years=2019 months=7-X weekdays=7-6 years=202a x=y)"
            ],
            CIB_HAS_BEEN_UPGRADED
            + (
                "Error: '2018-05-1X' is not a valid date value, use ISO 8601 date\n"
                "Error: Since '2019-03-05' is not sooner than until '2019-01-11'\n"
                "Error: '2019-05-0X' is not a valid date value, use ISO 8601 date\n"
                "Error: '3a' is not a valid months value, use a positive integer\n"
                "Error: invalid duration option 'x', allowed options are: "
                "'days', 'hours', 'minutes', 'months', 'seconds', 'weeks', "
                "'years'\n"
                "Error: Duplicate options in a single (sub)expression: 'months'\n"
                "Error: '202a' is not a valid years value, use an integer or "
                "integer-integer\n"
                "Error: '7-X' is not a valid months value, use 1..12 or 1..11-2..12\n"
                "Error: '7-6' is not a valid weekdays value, use 1..7 or 1..6-2..7\n"
                "Error: invalid datespec option 'x', allowed options are: "
                "'hours', 'minutes', 'monthdays', 'months', 'seconds', "
                "'weekdays', 'weeks', 'weekyears', 'yeardays', 'years'\n"
                "Error: Duplicate options in a single (sub)expression: 'years'\n"
                f"{self.meta_warning_not_validated}"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )


class RscDefaultsSetCreate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            etree.parse(cib).findall(".//rsc_defaults")[0]
        )
    ),
    DefaultsSetCreateMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]
    cib_tag = "rsc_defaults"
    meta_warning_not_validated = META_NOT_VALIDATED

    @skip_unless_pacemaker_supports_rsc_and_op_rules()
    def test_success_rules_rsc_op(self):
        score_present = not is_minimum_cib_schema_version(3, 9, 0)
        self.assert_effect(
            (
                self.cli_command
                + "set create id=X meta name1=value1 rule".split()
                + ["resource ::Dummy"]
            ),
            f"""\
            <{self.cib_tag}>
                <meta_attributes id="X">
                    <rule id="X-rule" boolean-op="and"
                        {'score="INFINITY"' if score_present else ""}
                    >
                        <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                    </rule>
                    <nvpair id="X-name1" name="name1" value="value1"/>
                </meta_attributes>
            </{self.cib_tag}>
            """,
            stderr_full=(
                CIB_HAS_BEEN_UPGRADED
                + fixture_warning_unknown_meta_primitive(["name1"])
                + DEFAULTS_MAY_BE_OVERRIDDEN
            ),
        )

    def test_node_attr_expressions(self):
        self.assert_pcs_fail(
            self.cli_command + ["set", "create", "rule", "defined attr"],
            CIB_HAS_BEEN_UPGRADED
            + (
                "Error: Keywords 'defined', 'not_defined', 'eq', 'ne', 'gte', "
                "'gt', 'lte' and 'lt' cannot be used in a rule in this command\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )


class OpDefaultsSetCreate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            etree.parse(cib).findall(".//op_defaults")[0]
        )
    ),
    DefaultsSetCreateMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]
    cib_tag = "op_defaults"

    def test_rule_error_messages(self):
        self.assert_pcs_fail(
            self.cli_command
            + [
                "set",
                "create",
                "rule",
                "defined attr1 or attr2 gte number 12a or "
                "attr3 lt version 3.2.1a or attr4 ne string test or attr5 lt 3 ",
            ],
            CIB_HAS_BEEN_UPGRADED
            + (
                "Error: '12a' is not a valid number attribute value, use a "
                "floating-point number\n"
                "Error: '3.2.1a' is not a valid version attribute value, use "
                "a version number (e.g. 1, 1.2, 1.23.45, ...)\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    @skip_unless_pacemaker_supports_rsc_and_op_rules()
    def test_success_rules_rsc_op(self):
        score_present = not is_minimum_cib_schema_version(3, 9, 0)
        self.assert_effect(
            self.cli_command
            + "-- set create id=X meta name1=value1 rule".split()
            + [
                "resource ::Dummy and (op start or op stop) and "
                "(defined attr1 or attr2 gte number -1.2 or "
                "attr3 lt version 3.2.1 or attr4 ne string test or attr5 lt 3) "
            ],
            f"""\
            <{self.cib_tag}>
                <meta_attributes id="X">
                    <rule id="X-rule" boolean-op="and"
                        {'score="INFINITY"' if score_present else ""}
                    >
                        <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                        <rule id="X-rule-rule" boolean-op="or"
                            {'score="0"' if score_present else ""}
                        >
                            <op_expression id="X-rule-rule-op-start"
                                name="start"
                            />
                            <op_expression id="X-rule-rule-op-stop"
                                name="stop"
                            />
                        </rule>
                        <rule id="X-rule-rule-1" boolean-op="or"
                            {'score="0"' if score_present else ""}
                        >
                            <expression id="X-rule-rule-1-expr"
                                operation="defined" attribute="attr1"
                            />
                            <expression id="X-rule-rule-1-expr-1"
                                attribute="attr2" operation="gte"
                                type="number" value="-1.2"
                            />
                            <expression id="X-rule-rule-1-expr-2"
                                attribute="attr3" operation="lt"
                                type="version" value="3.2.1"
                            />
                            <expression id="X-rule-rule-1-expr-3"
                                attribute="attr4" operation="ne"
                                type="string" value="test"
                            />
                            <expression id="X-rule-rule-1-expr-4"
                                attribute="attr5" operation="lt" value="3"
                            />
                        </rule>
                    </rule>
                    <nvpair id="X-name1" name="name1" value="value1"/>
                </meta_attributes>
            </{self.cib_tag}>
            """,
            stderr_full=(CIB_HAS_BEEN_UPGRADED + DEFAULTS_MAY_BE_OVERRIDDEN),
        )


class DefaultsSetDeleteMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []
    prefix = ""
    cib_tag = ""

    def setUp(self):
        super().setUp()
        xml_rsc = """
            <rsc_defaults>
                <meta_attributes id="rsc-set1" />
                <meta_attributes id="rsc-set2" />
                <meta_attributes id="rsc-set3" />
                <meta_attributes id="rsc-set4" />
            </rsc_defaults>
        """
        xml_op = """
            <op_defaults>
                <meta_attributes id="op-set1" />
                <meta_attributes id="op-set2" />
                <meta_attributes id="op-set3" />
                <meta_attributes id="op-set4" />
            </op_defaults>
        """
        xml_manip = XmlManipulation.from_file(empty_cib)
        xml_manip.append_to_first_tag_name("configuration", xml_rsc, xml_op)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

    def test_success(self):
        self.assert_effect(
            [
                self.cli_command
                + f"set delete {self.prefix}-set1 {self.prefix}-set3".split(),
                self.cli_command
                + f"set remove {self.prefix}-set1 {self.prefix}-set3".split(),
            ],
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="{self.prefix}-set2" />
                    <meta_attributes id="{self.prefix}-set4" />
                </{self.cib_tag}>
            """
            ),
        )


class RscDefaultsSetDelete(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            etree.parse(cib).findall(".//rsc_defaults")[0]
        )
    ),
    DefaultsSetDeleteMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]
    prefix = "rsc"
    cib_tag = "rsc_defaults"


class OpDefaultsSetDelete(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            etree.parse(cib).findall(".//op_defaults")[0]
        )
    ),
    DefaultsSetDeleteMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]
    prefix = "op"
    cib_tag = "op_defaults"


class DefaultsSetUpdateMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []
    prefix = ""
    cib_tag = ""
    meta_warning_not_validated = ""

    def test_success(self):
        xml = f"""
            <{self.cib_tag}>
                <meta_attributes id="my-set">
                    <nvpair id="my-set-name1" name="name1" value="value1" />
                    <nvpair id="my-set-name2" name="name2" value="value2" />
                    <nvpair id="my-set-name3" name="name3" value="value3" />
                </meta_attributes>
            </{self.cib_tag}>
        """
        xml_manip = XmlManipulation.from_file(empty_cib)
        xml_manip.append_to_first_tag_name("configuration", xml)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_effect(
            self.cli_command
            + "set update my-set meta name2=value2A name3=".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="my-set">
                        <nvpair id="my-set-name1" name="name1" value="value1" />
                        <nvpair id="my-set-name2" name="name2" value="value2A" />
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            stderr_full=(
                self.meta_warning_not_validated + DEFAULTS_MAY_BE_OVERRIDDEN
            ),
        )

        self.assert_effect(
            self.cli_command + "set update my-set meta name1= name2=".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="my-set" />
                </{self.cib_tag}>
            """
            ),
            stderr_full=(
                self.meta_warning_not_validated + DEFAULTS_MAY_BE_OVERRIDDEN
            ),
        )


class RscDefaultsSetUpdate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            etree.parse(cib).findall(".//rsc_defaults")[0]
        )
    ),
    DefaultsSetUpdateMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]
    prefix = "rsc"
    cib_tag = "rsc_defaults"
    meta_warning_not_validated = META_NOT_VALIDATED

    @skip_unless_pacemaker_supports_rsc_and_op_rules()
    def test_validate_meta_for_primitive(self):
        xml = f"""
            <{self.cib_tag}>
                <meta_attributes id="X">
                    <rule id="X-rule" boolean-op="and">
                        <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                    </rule>
                    <nvpair id="X-name1" name="name1" value="value1" />
                    <nvpair id="X-name2" name="name2" value="value2" />
                    <nvpair id="X-name3" name="name3" value="value3" />
                </meta_attributes>
            </{self.cib_tag}>
        """
        xml_manip = XmlManipulation.from_file(empty_cib_rules)
        xml_manip.append_to_first_tag_name("configuration", xml)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_effect(
            self.cli_command + "set update X meta name2=value2A name3=".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="X">
                        <rule id="X-rule" boolean-op="and">
                            <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                        </rule>
                        <nvpair id="X-name1" name="name1" value="value1" />
                        <nvpair id="X-name2" name="name2" value="value2A" />
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            stderr_full=(
                fixture_warning_unknown_meta_primitive(["name2"])
                + DEFAULTS_MAY_BE_OVERRIDDEN
            ),
        )


class OpDefaultsSetUpdate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            etree.parse(cib).findall(".//op_defaults")[0]
        )
    ),
    DefaultsSetUpdateMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]
    prefix = "op"
    cib_tag = "op_defaults"


class DefaultsSetUsageMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []

    def test_no_args(self):
        self.assert_pcs_fail(
            self.cli_command + ["set"],
            stderr_start=f"\nUsage: pcs {' '.join(self.cli_command)} set...\n",
        )

    def test_bad_command(self):
        self.assert_pcs_fail(
            self.cli_command + ["set", "bad-command"],
            stderr_start=f"\nUsage: pcs {' '.join(self.cli_command)} set ...\n",
        )


class RscDefaultsSetUsage(
    DefaultsSetUsageMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]


class OpDefaultsSetUsage(
    DefaultsSetUsageMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]


class DefaultsUpdateMixin(TestDefaultsMixin, AssertPcsMixin):
    cli_command = []
    prefix = ""
    cib_tag = ""
    meta_warning_not_validated = ""

    def test_success_legacy(self):
        write_file_to_tmpfile(empty_cib, self.temp_cib)

        command = self.cli_command[:]
        command.append("update")

        self.assert_effect(
            command + "name1=value1 name2=value2 name3=value3".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="{self.cib_tag}-meta_attributes">
                        <nvpair id="{self.cib_tag}-meta_attributes-name1"
                            name="name1" value="value1"
                        />
                        <nvpair id="{self.cib_tag}-meta_attributes-name2"
                            name="name2" value="value2"
                        />
                        <nvpair id="{self.cib_tag}-meta_attributes-name3"
                            name="name3" value="value3"
                        />
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            stderr_full=(
                DEFAULTS_MAY_BE_OVERRIDDEN + self.meta_warning_not_validated
            ),
        )

        self.assert_effect(
            command + "name2=value2A name3=".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="{self.cib_tag}-meta_attributes">
                        <nvpair id="{self.cib_tag}-meta_attributes-name1"
                            name="name1" value="value1"
                        />
                        <nvpair id="{self.cib_tag}-meta_attributes-name2"
                            name="name2" value="value2A"
                        />
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            stderr_full=(
                DEFAULTS_MAY_BE_OVERRIDDEN + self.meta_warning_not_validated
            ),
        )

        self.assert_effect(
            command + "name1= name2=".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="{self.cib_tag}-meta_attributes" />
                </{self.cib_tag}>
            """
            ),
            stderr_full=(
                DEFAULTS_MAY_BE_OVERRIDDEN + self.meta_warning_not_validated
            ),
        )


class RscDefaultsUpdate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            etree.parse(cib).findall(".//rsc_defaults")[0]
        )
    ),
    DefaultsUpdateMixin,
    TestCase,
):
    cli_command = ["resource", "defaults"]
    prefix = "rsc"
    cib_tag = "rsc_defaults"
    meta_warning_not_validated = META_NOT_VALIDATED

    @skip_unless_pacemaker_supports_rsc_and_op_rules()
    def test_legacy_validate_meta_for_primitive(self):
        xml = f"""
            <{self.cib_tag}>
                <meta_attributes id="X">
                    <rule id="X-rule" boolean-op="and">
                        <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                    </rule>
                    <nvpair id="X-name1" name="name1" value="value1" />
                    <nvpair id="X-name2" name="name2" value="value2" />
                    <nvpair id="X-name3" name="name3" value="value3" />
                </meta_attributes>
            </{self.cib_tag}>
        """
        xml_manip = XmlManipulation.from_file(empty_cib_rules)
        xml_manip.append_to_first_tag_name("configuration", xml)
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

        self.assert_effect(
            self.cli_command + "update name2=value2A name3=".split(),
            dedent(
                f"""\
                <{self.cib_tag}>
                    <meta_attributes id="X">
                        <rule id="X-rule" boolean-op="and">
                            <rsc_expression id="X-rule-rsc-Dummy" type="Dummy"/>
                        </rule>
                        <nvpair id="X-name1" name="name1" value="value1" />
                        <nvpair id="X-name2" name="name2" value="value2A" />
                    </meta_attributes>
                </{self.cib_tag}>
            """
            ),
            stderr_full=(
                DEFAULTS_MAY_BE_OVERRIDDEN
                + fixture_warning_unknown_meta_primitive(["name2"])
            ),
        )


class OpDefaultsUpdate(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            etree.parse(cib).findall(".//op_defaults")[0]
        )
    ),
    DefaultsUpdateMixin,
    TestCase,
):
    cli_command = ["resource", "op", "defaults"]
    prefix = "op"
    cib_tag = "op_defaults"



import logging
from unittest import TestCase, mock

from pcs.common import reports


class ReportProcessorToLog(TestCase):
    _REPORT_MESSAGE = reports.messages.ParseErrorCorosyncConf()
    _EXPECTED_MESSAGE = "Unable to parse corosync config"

    def setUp(self):
        self.mock_logger = mock.MagicMock(spec_set=logging.Logger)
        self.report_processor = reports.processor.ReportProcessorToLog(
            self.mock_logger
        )

    def test_log_error(self):
        self.report_processor.report(
            reports.ReportItem.error(self._REPORT_MESSAGE)
        )
        self.mock_logger.error.assert_called_once_with(self._EXPECTED_MESSAGE)
        self.mock_logger.warning.assert_not_called()
        self.mock_logger.info.assert_not_called()
        self.mock_logger.debug.assert_not_called()

    def test_log_warning(self):
        self.report_processor.report(
            reports.ReportItem.warning(self._REPORT_MESSAGE)
        )
        self.mock_logger.warning.assert_called_once_with(self._EXPECTED_MESSAGE)
        self.mock_logger.error.assert_not_called()
        self.mock_logger.info.assert_not_called()
        self.mock_logger.debug.assert_not_called()

    def test_log_deprecation(self):
        self.report_processor.report(
            reports.ReportItem.deprecation(self._REPORT_MESSAGE)
        )
        self.mock_logger.warning.assert_called_once_with(self._EXPECTED_MESSAGE)
        self.mock_logger.error.assert_not_called()
        self.mock_logger.info.assert_not_called()
        self.mock_logger.debug.assert_not_called()

    def test_log_info(self):
        self.report_processor.report(
            reports.ReportItem.info(self._REPORT_MESSAGE)
        )
        self.mock_logger.info.assert_called_once_with(self._EXPECTED_MESSAGE)
        self.mock_logger.error.assert_not_called()
        self.mock_logger.warning.assert_not_called()
        self.mock_logger.debug.assert_not_called()

    def test_log_debug(self):
        self.report_processor.report(
            reports.ReportItem.debug(self._REPORT_MESSAGE)
        )
        self.mock_logger.debug.assert_called_once_with(self._EXPECTED_MESSAGE)
        self.mock_logger.error.assert_not_called()
        self.mock_logger.warning.assert_not_called()
        self.mock_logger.info.assert_not_called()

    def test_log_with_context(self):
        self.report_processor.report(
            reports.ReportItem.error(
                self._REPORT_MESSAGE, context=reports.ReportItemContext("node")
            )
        )
        self.mock_logger.error.assert_called_once_with(
            f"node: {self._EXPECTED_MESSAGE}"
        )



from typing import (
    Any,
    Optional,
    cast,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    ArgsByKeywords,
    Argv,
    InputModifiers,
    group_by_keywords,
)
from pcs.common import reports
from pcs.common.resource_status import (
    EXACT_CHECK_STATES,
    InstancesQuantifierUnsupportedException,
    MembersQuantifierUnsupportedException,
    MoreChildrenQuantifierType,
    QueryException,
    ResourceException,
    ResourceNonExistentException,
    ResourceNotInGroupException,
    ResourcesStatusFacade,
    ResourceState,
    ResourceStateExactCheck,
    ResourceType,
    ResourceUnexpectedTypeException,
    can_be_promotable,
    can_be_unique,
)
from pcs.common.str_tools import (
    format_list,
    format_optional,
)


def _handle_query_result(result: bool, quiet: bool) -> SystemExit:
    if not quiet:
        print(result)

    if result:
        return SystemExit(0)
    return SystemExit(2)


def _handle_resource_exception(e: ResourceException) -> None:
    resource_id = f"{e.resource_id}{format_optional(e.instance_id, ':{}')}"
    if isinstance(e, ResourceNonExistentException):
        raise CmdLineInputError(f"Resource '{resource_id}' does not exist")
    if isinstance(e, ResourceNotInGroupException):
        raise CmdLineInputError(f"Resource '{resource_id}' is not in a group")
    if isinstance(e, ResourceUnexpectedTypeException):
        raise CmdLineInputError(
            (
                "Resource '{id}' has unexpected type '{real}'. This command "
                "works only for resources of type {expected}"
            ).format(
                id=resource_id,
                real=e.resource_type.value,
                expected=format_list(t.value for t in e.expected_types),
            )
        )
    raise CmdLineInputError(f"Unknown error with resource '{resource_id}'")


def _handle_query_exception(e: QueryException) -> None:
    if isinstance(e, MembersQuantifierUnsupportedException):
        raise CmdLineInputError(
            "'members' quantifier can be used only on group resources or "
            "group instances of cloned groups"
        )
    if isinstance(e, InstancesQuantifierUnsupportedException):
        raise CmdLineInputError(
            (
                "'instances' quantifier can be used only on clone resources "
                "and their instances, or on bundle resources and their replicas"
            )
        )
    raise CmdLineInputError(
        "Unknown error with the query", show_both_usage_and_message=True
    )


def _handle_is_modifiers(modifiers: InputModifiers) -> bool:
    modifiers.ensure_only_supported("--quiet", "-f")

    return modifiers.is_specified("--quiet")


def _handle_get_modifiers(modifiers: InputModifiers) -> None:
    modifiers.ensure_only_supported("-f")


def _get_resource_status_facade(lib: Any) -> ResourcesStatusFacade:
    dto = lib.status.resources_status()
    return ResourcesStatusFacade.from_resources_status_dto(dto)


def _parse_more_members_quantifier(
    sections: ArgsByKeywords,
    keyword: str,
) -> Optional[MoreChildrenQuantifierType]:
    if not sections.has_keyword(keyword):
        return None

    args = sections.get_args_flat(keyword)
    if len(args) != 1:
        raise CmdLineInputError()

    try:
        return MoreChildrenQuantifierType[args[0].upper()]
    except KeyError as e:
        raise CmdLineInputError(
            reports.messages.InvalidOptionValue(
                keyword,
                args[0],
                [
                    quantifier.name.lower()
                    for quantifier in MoreChildrenQuantifierType
                ],
            ).message
        ) from e


def _parse_expected_state(
    state_section: Argv,
) -> tuple[ResourceState, Optional[str]]:
    if not state_section or len(state_section) > 2:
        raise CmdLineInputError()

    try:
        expected_state = ResourceState[state_section[0].upper()]
    except KeyError as e:
        raise CmdLineInputError(
            reports.messages.InvalidOptionValue(
                "state",
                state_section[0],
                [state.name.lower() for state in ResourceState],
            ).message
        ) from e

    if len(state_section) == 1:
        return expected_state, None

    if expected_state not in EXACT_CHECK_STATES:
        raise CmdLineInputError()

    return expected_state, state_section[1]


def _pop_resource_id(argv: Argv) -> tuple[str, Optional[str]]:
    if len(argv) < 1:
        raise CmdLineInputError()
    resource_id = argv.pop(0)

    if not resource_id:
        raise CmdLineInputError()

    if ":" in resource_id:
        resource_id, instance_id = resource_id.rsplit(":", 1)
        return resource_id, instance_id

    return resource_id, None


def exists(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    resource_id, instance_id = _pop_resource_id(argv)
    if argv:
        raise CmdLineInputError()

    quiet = _handle_is_modifiers(modifiers)

    raise _handle_query_result(
        _get_resource_status_facade(lib).exists(resource_id, instance_id),
        quiet,
    )


def is_type(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    resource_id, instance_id = _pop_resource_id(argv)

    quiet = _handle_is_modifiers(modifiers)

    sections = group_by_keywords(argv, ["unique", "promotable"], "type")
    sections.ensure_unique_keywords()

    type_section = sections.get_args_flat("type")
    if len(type_section) != 1:
        raise CmdLineInputError()
    try:
        expected_type = ResourceType[type_section[0].upper()]
    except KeyError as e:
        raise CmdLineInputError(
            reports.messages.InvalidOptionValue(
                option_name="resource type",
                option_value=type_section[0],
                allowed_values=[
                    resource_type.value for resource_type in ResourceType
                ],
            ).message
        ) from e

    check_unique = sections.has_keyword("unique")
    if check_unique:
        if sections.get_args_flat("unique"):
            raise CmdLineInputError()
        if not can_be_unique(expected_type):
            raise CmdLineInputError(
                f"type '{expected_type.value}' cannot be unique"
            )

    check_promotable = sections.has_keyword("promotable")
    if check_promotable:
        if sections.get_args_flat("promotable"):
            raise CmdLineInputError()
        if not can_be_promotable(expected_type):
            raise CmdLineInputError(
                f"type '{expected_type.value}' cannot be promotable"
            )

    resources_status = _get_resource_status_facade(lib)
    try:
        result = (
            resources_status.get_type(resource_id, instance_id) == expected_type
        )
    except ResourceException as e:
        _handle_resource_exception(e)

    if result and check_unique:
        result = resources_status.is_unique(resource_id, instance_id)

    if result and check_promotable:
        result = resources_status.is_promotable(resource_id, instance_id)

    raise _handle_query_result(result, quiet)


def get_type(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    resource_id, instance_id = _pop_resource_id(argv)

    if argv:
        raise CmdLineInputError()

    _handle_get_modifiers(modifiers)
    resource_status = _get_resource_status_facade(lib)

    try:
        resource_type = resource_status.get_type(resource_id, instance_id)
    except ResourceException as e:
        _handle_resource_exception(e)

    output = [resource_type.value]
    if can_be_unique(resource_type) and resource_status.is_unique(
        resource_id, instance_id
    ):
        output.append("unique")
    if can_be_promotable(resource_type) and resource_status.is_promotable(
        resource_id, instance_id
    ):
        output.append("promotable")

    print(" ".join(output))


def is_stonith(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    resource_id, instance_id = _pop_resource_id(argv)
    if argv:
        raise CmdLineInputError()

    quiet = _handle_is_modifiers(modifiers)

    try:
        result = _get_resource_status_facade(lib).is_stonith(
            resource_id, instance_id
        )
    except ResourceException as e:
        _handle_resource_exception(e)

    raise _handle_query_result(result, quiet)


def get_members(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    resource_id, instance_id = _pop_resource_id(argv)
    if argv:
        raise CmdLineInputError()

    _handle_get_modifiers(modifiers)

    try:
        members = _get_resource_status_facade(lib).get_members(
            resource_id, instance_id
        )
    except ResourceException as e:
        _handle_resource_exception(e)

    print("\n".join(members))


def get_nodes(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    resource_id, instance_id = _pop_resource_id(argv)

    if argv:
        raise CmdLineInputError()

    _handle_get_modifiers(modifiers)
    try:
        nodes = _get_resource_status_facade(lib).get_nodes(
            resource_id, instance_id
        )
    except ResourceException as e:
        _handle_resource_exception(e)

    print("\n".join(nodes))


def is_state(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    # pylint: disable=too-many-locals
    resource_id, instance_id = _pop_resource_id(argv)

    sections = group_by_keywords(
        argv,
        {"on-node", "members", "instances"},
        implicit_first_keyword="state",
    )
    sections.ensure_unique_keywords()

    expected_state, expected_value = _parse_expected_state(
        sections.get_args_flat("state")
    )

    expected_node_name = None
    if sections.has_keyword("on-node"):
        node_section = sections.get_args_flat("on-node")
        if len(node_section) != 1:
            raise CmdLineInputError()
        expected_node_name = node_section[0]

    members_quantifier = _parse_more_members_quantifier(sections, "members")
    instances_quantifier = _parse_more_members_quantifier(sections, "instances")

    quiet = _handle_is_modifiers(modifiers)

    resource_status = _get_resource_status_facade(lib)
    try:
        if expected_value is not None and (
            expected_state in (ResourceState.LOCKED_TO, ResourceState.PENDING)
        ):
            result = resource_status.is_state_exact_value(
                resource_id,
                instance_id,
                cast(ResourceStateExactCheck, expected_state),
                expected_value,
                expected_node_name,
                members_quantifier,
                instances_quantifier,
            )
        else:
            result = resource_status.is_state(
                resource_id,
                instance_id,
                expected_state,
                expected_node_name,
                members_quantifier,
                instances_quantifier,
            )
    except ResourceException as e:
        _handle_resource_exception(e)
    except QueryException as e:
        _handle_query_exception(e)
    except NotImplementedError as e:
        raise CmdLineInputError(str(e)) from e

    raise _handle_query_result(result, quiet)


def _handle_is_in_container(
    real_id: Optional[str], expected_id: Optional[str], quiet: bool
) -> SystemExit:
    is_in_container = real_id is not None and (
        expected_id is None or real_id == expected_id
    )

    if not quiet:
        print(is_in_container)
        if real_id is not None:
            print(real_id)

    if not is_in_container:
        return SystemExit(2)
    return SystemExit(0)


def is_in_group(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    resource_id, instance_id = _pop_resource_id(argv)
    if len(argv) > 1:
        raise CmdLineInputError()
    expected_group_id = argv[0] if argv else None

    quiet = _handle_is_modifiers(modifiers)

    try:
        group_id = _get_resource_status_facade(lib).get_parent_group_id(
            resource_id, instance_id
        )
    except ResourceException as e:
        _handle_resource_exception(e)

    raise _handle_is_in_container(group_id, expected_group_id, quiet)


def is_in_clone(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    resource_id, instance_id = _pop_resource_id(argv)

    if len(argv) > 1:
        raise CmdLineInputError()
    expected_clone_id = argv[0] if argv else None

    quiet = _handle_is_modifiers(modifiers)

    try:
        clone_id = _get_resource_status_facade(lib).get_parent_clone_id(
            resource_id, instance_id
        )
    except ResourceException as e:
        _handle_resource_exception(e)

    raise _handle_is_in_container(clone_id, expected_clone_id, quiet)


def is_in_bundle(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    resource_id, instance_id = _pop_resource_id(argv)

    if len(argv) > 1:
        raise CmdLineInputError()
    expected_bundle_id = argv[0] if argv else None

    quiet = _handle_is_modifiers(modifiers)

    try:
        bundle_id = _get_resource_status_facade(lib).get_parent_bundle_id(
            resource_id, instance_id
        )
    except ResourceException as e:
        _handle_resource_exception(e)

    raise _handle_is_in_container(bundle_id, expected_bundle_id, quiet)


def get_index_in_group(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
        * -f - CIB file
        * --quiet - do not print anything to output
    """
    resource_id, instance_id = _pop_resource_id(argv)

    if argv:
        raise CmdLineInputError()

    _handle_get_modifiers(modifiers)

    try:
        index = _get_resource_status_facade(lib).get_index_in_group(
            resource_id, instance_id
        )
    except ResourceException as e:
        _handle_resource_exception(e)

    print(index)



from typing import Any, Mapping, Optional, Union

from lxml.etree import _Element

from pcs.common import const, reports
from pcs.common.pacemaker.defaults import CibDefaultsDto
from pcs.common.pacemaker.nvset import CibNvsetDto
from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.reports.item import ReportItem
from pcs.common.types import StringCollection
from pcs.lib.cib import const as cib_const
from pcs.lib.cib import nvpair_multi, sections
from pcs.lib.cib.rule import (
    RuleInEffectEval,
    RuleRoot,
    get_rule_evaluator,
    is_rsc_expressions_only,
    is_rsc_expressions_only_dto,
)
from pcs.lib.cib.tools import (
    IdProvider,
    get_pacemaker_version_by_which_cib_was_validated,
)
from pcs.lib.commands.resource import _validate_meta_attributes
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.resource_agent import const as ra_const


def resource_defaults_create(
    env: LibraryEnvironment,
    nvpairs: Mapping[str, str],
    nvset_options: Mapping[str, str],
    nvset_rule: Optional[str] = None,
    force_flags: reports.types.ForceFlags = (),
) -> None:
    """
    Create new resource defaults nvset

    env --
    nvpairs -- name-value pairs to be put into the new nvset
    nvset_options -- additional attributes of the created nvset
    nvset_rule -- optional rule describing when the created nvset applies
    force_flags -- list of flags codes
    """
    return _defaults_create(
        env,
        sections.RSC_DEFAULTS,
        dict(
            rule_allows_rsc_expr=True,
            rule_allows_op_expr=False,
            rule_allows_node_attr_expr=False,
        ),
        nvpairs,
        nvset_options,
        nvset_rule=nvset_rule,
        force_flags=force_flags,
    )


def operation_defaults_create(
    env: LibraryEnvironment,
    nvpairs: Mapping[str, str],
    nvset_options: Mapping[str, str],
    nvset_rule: Optional[str] = None,
    force_flags: reports.types.ForceFlags = (),
) -> None:
    """
    Create new operation defaults nvset

    env --
    nvpairs -- name-value pairs to be put into the new nvset
    nvset_options -- additional attributes of the created nvset
    nvset_rule -- optional rule describing when the created nvset applies
    force_flags -- list of flags codes
    """
    return _defaults_create(
        env,
        sections.OP_DEFAULTS,
        dict(
            rule_allows_rsc_expr=True,
            rule_allows_op_expr=True,
            rule_allows_node_attr_expr=True,
        ),
        nvpairs,
        nvset_options,
        nvset_rule=nvset_rule,
        force_flags=force_flags,
    )


def _defaults_create(
    env: LibraryEnvironment,
    cib_section_name: str,
    validator_options: Mapping[str, Any],
    nvpairs: Mapping[str, str],
    nvset_options: Mapping[str, str],
    nvset_rule: Optional[str] = None,
    force_flags: reports.types.ForceFlags = (),
) -> None:
    required_cib_version = None
    if nvset_rule:
        # Pacemaker 3 changed CIB schema for rules. We no longer support the
        # old schema, so we require CIB to be upgraded to the new one.
        required_cib_version = const.PCMK_RULES_PCMK3_SYNTAX_CIB_VERSION

    cib = env.get_cib(minimal_version=required_cib_version)
    id_provider = IdProvider(cib)
    cmd_runner = env.cmd_runner()

    validator = nvpair_multi.ValidateNvsetAppendNew(
        id_provider,
        nvpairs,
        nvset_options,
        nvset_rule=nvset_rule,
        **validator_options,
    )
    env.report_processor.report_list(
        validator.validate(force_options=reports.codes.FORCE in force_flags)
    )

    # meta-attrs validation - only for resource meta attributes
    if cib_section_name == sections.RSC_DEFAULTS:
        env.report_processor.report_list(
            __validate_meta_attrs_based_on_rule(
                cmd_runner, validator.get_parsed_rule(), nvpairs
            )
        )

    if env.report_processor.has_errors:
        raise LibraryError()

    nvpair_multi.nvset_append_new(
        sections.get(cib, cib_section_name),
        id_provider,
        get_pacemaker_version_by_which_cib_was_validated(cib),
        nvpair_multi.NVSET_META,
        nvpairs,
        nvset_options,
        nvset_rule=validator.get_parsed_rule(),
    )

    env.report_processor.report(
        ReportItem.warning(reports.messages.DefaultsCanBeOverridden())
    )
    env.push_cib()


def resource_defaults_config(
    env: LibraryEnvironment, evaluate_expired: bool
) -> CibDefaultsDto:
    """
    List all resource defaults nvsets

    env --
    evaluate_expired -- also evaluate whether rules are expired or in effect
    """
    cib = env.get_cib()
    rule_evaluator = get_rule_evaluator(
        cib, env.cmd_runner(), env.report_processor, evaluate_expired
    )

    def get_config(tag: nvpair_multi.NvsetTag) -> list[CibNvsetDto]:
        return _defaults_config(
            cib,
            tag,
            sections.RSC_DEFAULTS,
            rule_evaluator,
        )

    return CibDefaultsDto(
        instance_attributes=get_config(nvpair_multi.NVSET_INSTANCE),
        meta_attributes=get_config(nvpair_multi.NVSET_META),
    )


def operation_defaults_config(
    env: LibraryEnvironment, evaluate_expired: bool
) -> CibDefaultsDto:
    """
    List all operation defaults nvsets

    env --
    evaluate_expired -- also evaluate whether rules are expired or in effect
    """
    cib = env.get_cib()
    rule_evaluator = get_rule_evaluator(
        cib, env.cmd_runner(), env.report_processor, evaluate_expired
    )

    def get_config(tag: nvpair_multi.NvsetTag) -> list[CibNvsetDto]:
        return _defaults_config(
            cib,
            tag,
            sections.OP_DEFAULTS,
            rule_evaluator,
        )

    return CibDefaultsDto(
        instance_attributes=get_config(nvpair_multi.NVSET_INSTANCE),
        meta_attributes=get_config(nvpair_multi.NVSET_META),
    )


def _defaults_config(
    cib: _Element,
    nvset_tag: nvpair_multi.NvsetTag,
    cib_section_name: str,
    rule_evaluator: RuleInEffectEval,
) -> list[CibNvsetDto]:
    return [
        nvpair_multi.nvset_element_to_dto(nvset_el, rule_evaluator)
        for nvset_el in nvpair_multi.find_nvsets(
            sections.get(cib, cib_section_name), nvset_tag
        )
    ]


def resource_defaults_remove(
    env: LibraryEnvironment, nvset_id_list: StringCollection
) -> None:
    """
    Remove specified resource defaults nvsets

    env --
    nvset_id_list -- nvset IDs to be removed
    """
    return _defaults_remove(env, sections.RSC_DEFAULTS, nvset_id_list)


def operation_defaults_remove(
    env: LibraryEnvironment, nvset_id_list: StringCollection
) -> None:
    """
    Remove specified operation defaults nvsets

    env --
    nvset_id_list -- nvset IDs to be removed
    """
    return _defaults_remove(env, sections.OP_DEFAULTS, nvset_id_list)


def _defaults_remove(
    env: LibraryEnvironment,
    cib_section_name: str,
    nvset_id_list: StringCollection,
) -> None:
    if not nvset_id_list:
        return
    nvset_elements, report_list = nvpair_multi.find_nvsets_by_ids(
        sections.get(env.get_cib(), cib_section_name), nvset_id_list
    )
    if env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()
    nvpair_multi.nvset_remove(nvset_elements)
    env.push_cib()


def resource_defaults_update(
    env: LibraryEnvironment,
    nvset_id: Optional[str],
    nvpairs: Mapping[str, str],
) -> None:
    """
    Update specified resource defaults nvset

    env --
    nvset_id -- nvset ID to be updated; if None, update an existing nvset if
        there is only one
    nvpairs -- name-value pairs to be put into the nvset
    """
    return _defaults_update(
        env,
        sections.RSC_DEFAULTS,
        nvset_id,
        nvpairs,
        reports.const.PCS_COMMAND_RESOURCE_DEFAULTS_UPDATE,
    )


def operation_defaults_update(
    env: LibraryEnvironment,
    nvset_id: Optional[str],
    nvpairs: Mapping[str, str],
) -> None:
    """
    Update specified operation defaults nvset

    env --
    nvset_id -- nvset ID to be updated; if None, update an existing nvset if
        there is only one
    nvpairs -- name-value pairs to be put into the nvset
    """
    return _defaults_update(
        env,
        sections.OP_DEFAULTS,
        nvset_id,
        nvpairs,
        reports.const.PCS_COMMAND_OPERATION_DEFAULTS_UPDATE,
    )


def _defaults_update(
    env: LibraryEnvironment,
    cib_section_name: str,
    nvset_id: Optional[str],
    nvpairs: Mapping[str, str],
    pcs_command: reports.types.PcsCommand,
) -> None:
    # ruff: noqa: PLR0912 Too many branches
    cib = env.get_cib()
    id_provider = IdProvider(cib)
    cmd_runner = env.cmd_runner()

    if nvset_id is None:
        # Backward compatibility code to support an old use case where no id
        # was requested and provided and the first meta_attributes nvset was
        # created / updated. However, we check that there is only one nvset
        # present in the CIB to prevent breaking the configuration with
        # multiple nvsets in place.

        # This is to be supported as it provides means of easily managing
        # defaults if only one set of defaults is needed.

        # TODO move this to a separate lib command.

        if not nvpairs:
            return

        # Do not create new defaults element if we are only removing values
        # from it.
        only_removing = True
        for value in nvpairs.values():
            if value != "":
                only_removing = False
                break
        if only_removing and not sections.exists(cib, cib_section_name):
            env.report_processor.report(
                ReportItem.warning(reports.messages.DefaultsCanBeOverridden())
            )
            return

        nvset_elements = nvpair_multi.find_nvsets(
            sections.get(cib, cib_section_name), nvpair_multi.NVSET_META
        )
        if len(nvset_elements) > 1:
            env.report_processor.report(
                reports.item.ReportItem.error(
                    reports.messages.CibNvsetAmbiguousProvideNvsetId(
                        pcs_command
                    )
                )
            )
            raise LibraryError()
        env.report_processor.report(
            ReportItem.warning(reports.messages.DefaultsCanBeOverridden())
        )
        if len(nvset_elements) == 1:
            # meta-attrs validation - only for resource meta attributes
            if cib_section_name == sections.RSC_DEFAULTS:
                nvset_dto = nvpair_multi.nvset_element_to_dto(
                    nvset_elements[0],
                    get_rule_evaluator(
                        cib,
                        cmd_runner,
                        env.report_processor,
                        evaluate_expired=False,
                    ),
                )
                env.report_processor.report_list(
                    __validate_meta_attrs_based_on_rule(
                        cmd_runner, nvset_dto.rule, nvpairs
                    )
                )
                if env.report_processor.has_errors:
                    raise LibraryError()

            nvpair_multi.nvset_update(nvset_elements[0], id_provider, nvpairs)
        elif only_removing:
            # do not create new nvset if there is none and we are only removing
            # nvpairs
            return
        else:
            # meta-attrs validation - only for resource meta attributes
            if cib_section_name == sections.RSC_DEFAULTS:
                env.report_processor.report_list(
                    __validate_meta_attrs_based_on_rule(
                        cmd_runner, None, nvpairs
                    )
                )
                if env.report_processor.has_errors:
                    raise LibraryError()

            nvpair_multi.nvset_append_new(
                sections.get(cib, cib_section_name),
                id_provider,
                get_pacemaker_version_by_which_cib_was_validated(cib),
                nvpair_multi.NVSET_META,
                nvpairs,
                {},
            )
        env.push_cib()
        return

    nvset_element_list, report_list = nvpair_multi.find_nvsets_by_ids(
        sections.get(cib, cib_section_name), [nvset_id]
    )
    if env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()

    nvset_element = nvset_element_list[0]

    # meta-attrs validation - only for resource meta attributes
    if cib_section_name == sections.RSC_DEFAULTS:
        nvset_dto = nvpair_multi.nvset_element_to_dto(
            nvset_element,
            get_rule_evaluator(
                cib, cmd_runner, env.report_processor, evaluate_expired=False
            ),
        )
        env.report_processor.report_list(
            __validate_meta_attrs_based_on_rule(
                cmd_runner, nvset_dto.rule, nvpairs
            )
        )

    if env.report_processor.has_errors:
        raise LibraryError()

    nvpair_multi.nvset_update(nvset_element, id_provider, nvpairs)
    env.report_processor.report(
        ReportItem.warning(reports.messages.DefaultsCanBeOverridden())
    )
    env.push_cib()


def __validate_meta_attrs_based_on_rule(
    cmd_runner: CommandRunner,
    rule: Union[None, RuleRoot, CibRuleExpressionDto],
    meta_attrs: Mapping[str, str],
) -> reports.ReportItemList:
    # Currently, pcmk provides meta-attrs definition for primitive and stonith
    # only. That is not sufficient - meta-attrs valid for other resource types
    # would be rejected by pcs. So we only validate when we are sure the
    # defaults apply only to a primitive or a stonith.
    # TODO Once pcmk provides meta-attrs definition for all resource types, run
    # the validation in all cases.
    if not meta_attrs:
        return []

    if rule and (
        (isinstance(rule, RuleRoot) and is_rsc_expressions_only(rule))
        or (
            isinstance(rule, CibRuleExpressionDto)
            and is_rsc_expressions_only_dto(rule)
        )
    ):
        # Rules with resource expressions only are easily detectable and they
        # apply to primitive resources only.
        return _validate_meta_attributes(
            cmd_runner,
            [ra_const.PRIMITIVE_META, ra_const.STONITH_META],
            meta_attrs,
        )
    # Meta-attrs in nvset with other rules or no rules cannot be validated,
    # they may apply to all resource types and we don't have meta-attrs
    # definition for those.
    return [
        reports.ReportItem.warning(
            reports.messages.MetaAttrsNotValidatedUnsupportedType(
                [
                    cib_const.TAG_RESOURCE_BUNDLE,
                    cib_const.TAG_RESOURCE_CLONE,
                    cib_const.TAG_RESOURCE_GROUP,
                    cib_const.TAG_RESOURCE_PRIMITIVE,
                ]
            )
        )
    ]



import multiprocessing as mp
import os
import signal
import socket
import sys
from logging import Logger
from pathlib import Path
from typing import (
    Iterable,
    Optional,
)

from tornado.ioloop import (
    IOLoop,
    PeriodicCallback,
)
from tornado.locks import Lock
from tornado.web import Application

from pcs import settings
from pcs.common import capabilities
from pcs.common.types import StringCollection
from pcs.daemon import (
    log,
    ruby_pcsd,
    ssl,
    systemd,
)
from pcs.daemon.app import (
    api_v0,
    api_v1,
    api_v2,
    auth,
    sinatra_remote,
    sinatra_ui,
)
from pcs.daemon.app import capabilities as capabilities_app

try:
    from pcs.daemon.app import webui
except ImportError:
    webui = None

from pcs.common.communication.logger import CommunicatorLogger
from pcs.common.node_communicator import NodeCommunicatorFactory
from pcs.common.reports.processor import ReportProcessorToLog
from pcs.daemon.app.common import (
    Http404Handler,
    RedirectHandler,
)
from pcs.daemon.async_tasks.scheduler import (
    Scheduler,
    SchedulerConfig,
)
from pcs.daemon.async_tasks.task import TaskConfig
from pcs.daemon.env import prepare_env
from pcs.daemon.http_server import HttpsServerManage
from pcs.daemon.pcs_cfgsync import CfgSyncPullManager
from pcs.lib.auth.provider import AuthProvider


class SignalInfo:
    async_scheduler: Optional[Scheduler] = None
    server_manage = None
    ioloop_started = False


def handle_signal(incoming_signal, frame):
    del frame
    log.pcsd.warning("Caught signal: %s, shutting down", incoming_signal)
    if SignalInfo.server_manage:
        SignalInfo.server_manage.stop()
    if SignalInfo.async_scheduler:
        SignalInfo.async_scheduler.terminate_nowait()
    if SignalInfo.ioloop_started:
        IOLoop.current().stop()
    raise SystemExit(0)


def sign_ioloop_started():
    SignalInfo.ioloop_started = True


def create_pull_manager(logger: Logger) -> CfgSyncPullManager:
    log_report_processor = ReportProcessorToLog(logger)
    node_communicator = NodeCommunicatorFactory(
        CommunicatorLogger([log_report_processor]),
        user=None,
        groups=None,
        # 30, because cfgsync in ruby used the same value. Probably
        # because 30 was default for request timeouts back then.
        # This value might be reconsider
        request_timeout=30,
    ).get_communicator()
    return CfgSyncPullManager(log_report_processor, node_communicator, logger)


async def config_sync(
    sync_config_lock: Lock, config_puller: CfgSyncPullManager
):
    async with sync_config_lock:
        # run_cfgsync sends requests to all cluster nodes, including the local
        # one. However, this runs in the same IOLoop as the async_scheduler
        # that handles these requests -> so run cfgsync waits for responses
        # but the async scheduler cannot process them, resulting in timeout.
        # So we need to run this blocking function in executor
        next_run_after_seconds = await IOLoop.current().run_in_executor(
            None, lambda: config_puller.run_cfgsync()
        )
    IOLoop.current().call_later(
        next_run_after_seconds,
        lambda: config_sync(sync_config_lock, config_puller),
    )


def configure_app(  # noqa: PLR0913
    async_scheduler: Scheduler,
    auth_provider: AuthProvider,
    session_lifetime: int,
    ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
    sync_config_lock: Lock,
    webui_dir: str,
    webui_fallback: str,
    pcsd_capabilities: Iterable[capabilities.Capability],
    *,
    debug: bool = False,
):
    # pylint: disable=too-many-arguments
    def make_app(https_server_manage: HttpsServerManage):
        """
        https_server_manage -- allows to control the server (specifically
            reload its SSL certificates). A relevant handler should get this
            object via the method `initialize`.
        """

        routes = api_v2.get_routes(async_scheduler, auth_provider)
        routes.extend(api_v1.get_routes(async_scheduler, auth_provider))
        routes.extend(api_v0.get_routes(async_scheduler, auth_provider))
        routes.extend(auth.get_routes(auth_provider))
        routes.extend(
            capabilities_app.get_routes(auth_provider, pcsd_capabilities)
        )
        routes.extend(
            sinatra_remote.get_routes(
                ruby_pcsd_wrapper,
                sync_config_lock,
                https_server_manage,
                auth_provider,
            )
        )

        if webui:
            session_storage = webui.session.Storage(session_lifetime)
            routes.extend(
                [(r"/(ui)?", RedirectHandler, dict(url="/ui/"))]
                + webui.core.get_routes(
                    url_prefix="/ui/",
                    app_dir=webui_dir,
                    fallback_page_path=webui_fallback,
                    session_storage=session_storage,
                    auth_provider=auth_provider,
                )
                + webui.sinatra_ui.get_routes(
                    session_storage, auth_provider, ruby_pcsd_wrapper
                )
            )
        else:
            # Even with disabled (standalone) webui the following routes must be
            # provided because they can be used via unix socket from cockpit.
            routes.extend(
                sinatra_ui.get_routes(auth_provider, ruby_pcsd_wrapper)
            )

        return Application(
            routes, debug=debug, default_handler_class=Http404Handler
        )

    return make_app


def _print_version(argv: StringCollection) -> None:
    try:
        print(settings.pcs_version)
        if "--full" in argv:
            print(
                capabilities.capabilities_to_codes_str(
                    capabilities.get_pcsd_capabilities()
                )
            )
    except capabilities.CapabilitiesError as e:
        sys.stderr.write(f"Error: {e.msg}\n")
        raise SystemExit(1) from e


def main(argv=None) -> None:
    # set the way how processes are started
    # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
    # avoid deadlock in multiprocessing.pool.Pool on terminate
    # https://github.com/python/cpython/issues/73945
    mp.set_start_method(method="forkserver")

    argv = argv if argv is not None else sys.argv[1:]
    if "--version" in argv:
        _print_version(argv)
        return

    if argv:
        sys.stderr.write(f"Error: option {argv[0]} is not recognized\n")
        raise SystemExit(1)

    signal.signal(signal.SIGTERM, handle_signal)
    signal.signal(signal.SIGINT, handle_signal)

    Path(settings.pcsd_log_location).touch(mode=0o600, exist_ok=True)
    log.setup(settings.pcsd_log_location)

    env = prepare_env(os.environ, log.pcsd)
    if env.has_errors:
        raise SystemExit(1)

    if env.PCSD_DEBUG:
        log.enable_debug()

    async_scheduler = Scheduler(
        SchedulerConfig(
            worker_count=env.PCSD_WORKER_COUNT,
            max_worker_count=env.PCSD_MAX_WORKER_COUNT,
            worker_reset_limit=env.PCSD_WORKER_RESET_LIMIT,
            deadlock_threshold_timeout=env.PCSD_DEADLOCK_THRESHOLD_TIMEOUT,
            task_config=TaskConfig(
                abandoned_timeout=env.PCSD_TASK_ABANDONED_TIMEOUT,
                unresponsive_timeout=env.PCSD_TASK_UNRESPONSIVE_TIMEOUT,
                deletion_timeout=env.PCSD_TASK_DELETION_TIMEOUT,
            ),
        )
    )
    auth_provider = AuthProvider(log.pcsd)
    SignalInfo.async_scheduler = async_scheduler

    sync_config_lock = Lock()
    ruby_pcsd_wrapper = ruby_pcsd.Wrapper(
        settings.pcsd_ruby_socket,
        debug=env.PCSD_DEBUG,
    )

    try:
        pcsd_capabilities = capabilities.get_pcsd_capabilities()
    except capabilities.CapabilitiesError as e:
        pcsd_capabilities = []
        log.pcsd.error(e.msg)

    make_app = configure_app(
        async_scheduler,
        auth_provider,
        env.PCSD_SESSION_LIFETIME,
        ruby_pcsd_wrapper,
        sync_config_lock,
        env.WEBUI_DIR,
        env.WEBUI_FALLBACK,
        pcsd_capabilities,
        debug=env.PCSD_DEV,
    )
    pcsd_ssl = ssl.PcsdSSL(
        server_name=socket.gethostname(),
        cert_location=settings.pcsd_cert_location,
        key_location=settings.pcsd_key_location,
        ssl_options=env.PCSD_SSL_OPTIONS,
        ssl_ciphers=env.PCSD_SSL_CIPHERS,
    )
    try:
        SignalInfo.server_manage = HttpsServerManage(
            make_app,
            port=env.PCSD_PORT,
            bind_addresses=env.PCSD_BIND_ADDR,
            ssl=pcsd_ssl,
            unix_socket_path=settings.pcsd_unix_socket,
        ).start()
    except socket.gaierror as e:
        log.pcsd.error(
            "Unable to bind to specific address(es), exiting: %s ", e
        )
        raise SystemExit(1) from e
    except OSError as e:
        log.pcsd.error("Unable to start pcsd daemon, exiting: %s ", e)
        raise SystemExit(1) from e
    except ssl.SSLCertKeyException as e:
        for error in e.args:
            log.pcsd.error(error)
        log.pcsd.error("Invalid SSL certificate and/or key, exiting")
        raise SystemExit(1) from e

    PeriodicCallback(
        async_scheduler.perform_actions,
        callback_time=env.PCSD_CHECK_INTERVAL_MS,
    ).start()
    ioloop = IOLoop.current()
    ioloop.add_callback(sign_ioloop_started)
    if systemd.is_systemd() and env.NOTIFY_SOCKET:
        ioloop.add_callback(systemd.notify, env.NOTIFY_SOCKET)

    cfgsync_pull_manager = create_pull_manager(log.pcsd)
    ioloop.add_callback(
        lambda: config_sync(sync_config_lock, cfgsync_pull_manager)
    )
    ioloop.start()



import os.path

BIN_MOCK_DIR = os.path.dirname(os.path.abspath(__file__))

CRM_RESOURCE_BIN = os.path.abspath(
    os.path.join(BIN_MOCK_DIR, "pcmk/crm_resource")
)
PACEMAKER_FENCED_BIN = os.path.abspath(
    os.path.join(BIN_MOCK_DIR, "pcmk/pacemaker-fenced")
)
STONITH_ADMIN_BIN = os.path.abspath(
    os.path.join(BIN_MOCK_DIR, "pcmk/stonith_admin")
)

MOCK_SETTINGS = {
    "crm_resource_exec": CRM_RESOURCE_BIN,
    "pacemaker_fenced_exec": PACEMAKER_FENCED_BIN,
    "stonith_admin_exec": STONITH_ADMIN_BIN,
}


def get_mock_settings(*required_settings):
    if not required_settings:
        required_settings = MOCK_SETTINGS.keys()
    return {
        key: value
        for key, value in MOCK_SETTINGS.items()
        if key in required_settings
    }



import json
from unittest import TestCase, mock

from pcs import settings
from pcs.common import reports
from pcs.lib.cib.element_description import TAG_LIST_SUPPORTS_DESCRIPTION
from pcs.lib.commands import cib as lib

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.fixture_cib import modify_cib
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import read_test_resource


def _constraints(*argv):
    return f"<constraints>{''.join(argv)}</constraints>"


FIXTURE_LOC_CONSTRAINT_WITH_1_RULE = """
    <rsc_location id="lr1" rsc="A">
        <rule id="r1" boolean-op="and" score="100">
            <expression id="r1e1" operation="eq"
                attribute="#uname" value="node1"/>
            <date_expression id="r1e2" operation="gt"
                start="1970-01-01"/>
        </rule>
    </rsc_location>
"""

FIXTURE_LOC_CONSTRAINT_WITH_2_RULES = """
    <rsc_location id="lr2" rsc="B">
        <rule id="r2" score="-INFINITY" >
            <expression id="r2e1" attribute="pingd"
                operation="lt" value="3000"/>
        </rule>
        <rule id="r3" score-attribute="pingd" >
            <expression id="r3e1" attribute="pingd"
                operation="defined"/>
        </rule>
    </rsc_location>
"""

FIXTURE_TWO_LOC_CONSTRAINTS_WITH_RULES = _constraints(
    FIXTURE_LOC_CONSTRAINT_WITH_1_RULE,
    FIXTURE_LOC_CONSTRAINT_WITH_2_RULES,
)

EXPECTED_TYPES_FOR_REMOVE = ["constraint", "location rule", "resource"]


def fixture_remote_resource(resource_id: str) -> str:
    return f'<primitive id="{resource_id}" class="ocf" type="remote" provider="pacemaker"/>'


def fixture_guest_resource(resource_id: str) -> str:
    return f"""
        <primitive id="{resource_id}" class="heartbeat" type="VirtualDomain">
            <meta_attributes id="meta">
                <nvpair id="meta-remote-node" name="remote-node" value="remote"/>
            </meta_attributes>
        </primitive>
    """


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_rng/api/api-result.rng")
)
class RemoveElements(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_ids_not_found_and_unsupported_types(self):
        self.config.runner.cib.load(
            tags="""
                <tags>
                    <tag id="T">
                        <obj_ref id="A"/>
                    </tag>
                </tags>
            """
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_elements(
                self.env_assist.get_env(), ["A", "T", "C"]
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_found("A", expected_types=[]),
                fixture.report_not_found("C", expected_types=[]),
                fixture.report_unexpected_element(
                    "T", "tag", EXPECTED_TYPES_FOR_REMOVE
                ),
            ]
        )

    def test_duplicate_ids_specified(self):
        self.config.runner.cib.load(
            constraints="""
                <constraints>
                    <rsc_location id="l1" rsc="A" node="node1" score="200"/>
                    <rsc_location id="l2" rsc="A" node="node2" score="100"/>
                </constraints>
            """
        )
        self.config.env.push_cib(
            constraints="""
                <constraints>
                    <rsc_location id="l2" rsc="A" node="node2" score="100"/>
                </constraints>
            """
        )
        lib.remove_elements(self.env_assist.get_env(), ["l1", "l1"])

    def test_remove_constraints(self):
        self.config.runner.cib.load(
            constraints="""
                <constraints>
                    <rsc_location id="l1" rsc="A" node="node1" score="200"/>
                    <rsc_location id="l2" rsc="A" node="node2" score="100"/>
                    <rsc_order id="o1" first="A" then="B"/>
                    <rsc_order id="o2" first="A" then="B"/>
                    <rsc_colocation id="c1" rsc="A" with-rsc="B" score="100"/>
                    <rsc_colocation id="c2" rsc="A" with-rsc="B" score="100"/>
                    <rsc_ticket id="t1" ticket="T" rsc="A"/>
                    <rsc_ticket id="t2" ticket="T" rsc="B"/>
                </constraints>
            """
        )
        self.config.env.push_cib(
            constraints="""
                <constraints>
                    <rsc_location id="l2" rsc="A" node="node2" score="100"/>
                    <rsc_order id="o1" first="A" then="B"/>
                    <rsc_ticket id="t2" ticket="T" rsc="B"/>
                </constraints>
            """
        )
        lib.remove_elements(
            self.env_assist.get_env(), ["l1", "o2", "c1", "c2", "t1"]
        )

    def test_remove_location_rules(self):
        self.config.runner.cib.load(
            constraints=FIXTURE_TWO_LOC_CONSTRAINTS_WITH_RULES
        )
        self.config.env.push_cib(constraints="<constraints/>")
        lib.remove_elements(self.env_assist.get_env(), ["r1", "r2", "r3"])
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.CIB_REMOVE_DEPENDANT_ELEMENTS,
                    id_tag_map={"lr1": "rsc_location", "lr2": "rsc_location"},
                ),
            ]
        )

    def test_remove_location_rule_expressions(self):
        self.config.runner.cib.load(
            constraints=FIXTURE_TWO_LOC_CONSTRAINTS_WITH_RULES
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_elements(
                self.env_assist.get_env(), ["r1e1", "r1e2"]
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.report_unexpected_element(
                    "r1e1", "expression", EXPECTED_TYPES_FOR_REMOVE
                ),
                fixture.report_unexpected_element(
                    "r1e2", "date_expression", EXPECTED_TYPES_FOR_REMOVE
                ),
            ]
        )

    def test_remove_resources(self):
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="P-1"/>
                    <primitive id="P-2"/>
                    <group id="G">
                        <primitive id="G-1"/>
                        <primitive id="G-2"/>
                    </group>
                    <clone id="C">
                        <group id="C-G">
                            <primitive id="C-G-1"/>
                            <primitive id="C-G-2"/>
                        </group>
                    </clone>
                </resources>
            """
        )
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <resource id="P-1" managed="true" role="Stopped"/>
                    <group id="G" number_resources="2" maintenance="false" managed="true" disabled="false">
                        <resource id="G-1" managed="true" role="Stopped"/>
                        <resource id="G-2" managed="true" role="Stopped"/>
                    </group>
                    <clone id="C" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false">
                        <group id="C-G" number_resources="2" maintenance="false" managed="true" disabled="false">
                            <resource id="C-G-1" managed="true" role="Stopped"/>
                            <resource id="C-G-2" managed="true" role="Stopped"/>
                        </group>
                    </clone>
                </resources>
            """
        )
        self.config.env.push_cib(
            resources="""
                <resources>
                    <primitive id="P-2"/>
                </resources>
            """
        )
        lib.remove_elements(
            self.env_assist.get_env(),
            ["P-1", "G", "C"],
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.CIB_REMOVE_DEPENDANT_ELEMENTS,
                    id_tag_map={
                        "G-1": "primitive",
                        "G-2": "primitive",
                        "C-G": "group",
                        "C-G-1": "primitive",
                        "C-G-2": "primitive",
                    },
                )
            ]
        )

    def test_remove_resource_guest(self):
        cib = modify_cib(
            read_test_resource("cib-empty.xml"),
            resources=f"""
                <resources>
                    {fixture_guest_resource("R1")}
                    <group id="G1">
                        {fixture_guest_resource("R2")}
                    </group>
                    <clone id="C1">
                        {fixture_guest_resource("R3")}
                    </clone>
                    <clone id="C2">
                        <group id="G2">
                            {fixture_guest_resource("R4")}
                        </group>
                    </clone>
                </resources>
            """,
        )
        self.config.runner.cib.load_content(cib)
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <resource id="R1" managed="true" role="Stopped"/>
                    <group id="G1" number_resources="1" maintenance="false" managed="true" disabled="false">
                        <resource id="R2" managed="true" role="Stopped"/>
                    </group>
                    <clone id="C1" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false">
                        <resource id="R3" managed="true" role="Stopped"/>
                    </clone>
                    <clone id="C1" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false">
                        <group id="G2" number_resources="1" maintenance="false" managed="true" disabled="false">
                            <resource id="R4" managed="true" role="Stopped"/>
                        </group>
                    </clone>
                </resources>
            """
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_elements(
                self.env_assist.get_env(), ["R1", "R2", "R3", "R4"]
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_REMOVE_GUEST,
                    resource_id="R1",
                ),
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_REMOVE_GUEST,
                    resource_id="R2",
                ),
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_REMOVE_GUEST,
                    resource_id="R3",
                ),
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_REMOVE_GUEST,
                    resource_id="R4",
                ),
            ]
        )

    def test_remove_resource_remote(self):
        cib = modify_cib(
            read_test_resource("cib-empty.xml"),
            resources=f"""
                <resources>
                    {fixture_remote_resource("R1")}
                    <group id="G1">
                        {fixture_remote_resource("R2")}
                    </group>
                    <clone id="C1">
                        {fixture_remote_resource("R3")}
                    </clone>
                    <clone id="C2">
                        <group id="G2">
                            {fixture_remote_resource("R4")}
                        </group>
                    </clone>
                </resources>
            """,
        )
        self.config.runner.cib.load_content(cib)
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <resource id="R1" managed="true" role="Stopped"/>
                    <group id="G1" number_resources="1" maintenance="false" managed="true" disabled="false">
                        <resource id="R2" managed="true" role="Stopped"/>
                    </group>
                    <clone id="C1" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false">
                        <resource id="R3" managed="true" role="Stopped"/>
                    </clone>
                    <clone id="C1" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false">
                        <group id="G2" number_resources="1" maintenance="false" managed="true" disabled="false">
                            <resource id="R4" managed="true" role="Stopped"/>
                        </group>
                    </clone>
                </resources>
            """
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_elements(
                self.env_assist.get_env(), ["R1", "R2", "R3", "R4"]
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_REMOVE_REMOTE,
                    resource_id="R1",
                ),
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_REMOVE_REMOTE,
                    resource_id="R2",
                ),
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_REMOVE_REMOTE,
                    resource_id="R3",
                ),
                fixture.error(
                    reports.codes.USE_COMMAND_NODE_REMOVE_REMOTE,
                    resource_id="R4",
                ),
            ]
        )

    def test_remove_resource_multiple_dependencies(self):
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <primitive id="B"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_location id="L1" rsc="A" node="node1" score="200"/>
                    <rsc_location id="L2" rsc="B" node="node1" score="200"/>
                    <rsc_colocation score="-1" id="C">
                        <resource_set role="Started" id="SET">
                            <resource_ref id="A"/>
                            <resource_ref id="B"/>
                        </resource_set>
                    </rsc_colocation>
                </constraints>
            """,
            tags="""
                <tags>
                    <tag id="T">
                        <obj_ref id="A"/>
                        <obj_ref id="B"/>
                    </tag>
                </tags>
            """,
            fencing_topology="""
                <fencing-topology>
                    <fencing-level index="1" devices="A" target="NODE-A" id="FL"/>
                </fencing-topology>
            """,
            optional_in_conf="""
                <acls>
                    <acl_role id="ROLE">
                        <acl_permission id="PERMISSION" kind="read" reference="A"/>
                    </acl_role>
                    <acl_target id="TARGET">
                        <role id="ROLE"/>
                    </acl_target>
                </acls>
            """,
        )
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <resource id="A" role="Stopped" managed="true"/>
                    <resource id="B" role="Stopped" managed="true"/>
                </resources>
            """
        )
        self.config.env.push_cib(
            resources="""
                <resources>
                    <primitive id="B"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_location id="L2" rsc="B" node="node1" score="200"/>
                    <rsc_colocation score="-1" id="C">
                        <resource_set role="Started" id="SET">
                            <resource_ref id="B"/>
                        </resource_set>
                    </rsc_colocation>
                </constraints>
            """,
            tags="""
                <tags>
                    <tag id="T">
                        <obj_ref id="B"/>
                    </tag>
                </tags>
            """,
            fencing_topology="<fencing-topology/>",
            optional_in_conf="""
                <acls>
                    <acl_role id="ROLE">
                    </acl_role>
                    <acl_target id="TARGET">
                        <role id="ROLE"/>
                    </acl_target>
                </acls>
            """,
        )
        lib.remove_elements(self.env_assist.get_env(), ["A"])
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.CIB_REMOVE_DEPENDANT_ELEMENTS,
                    id_tag_map={
                        "L1": "rsc_location",
                        "FL": "fencing-level",
                        "PERMISSION": "acl_permission",
                    },
                ),
                fixture.info(
                    reports.codes.CIB_REMOVE_REFERENCES,
                    id_tag_map={
                        "A": "primitive",
                        "T": "tag",
                        "PERMISSION": "acl_permission",
                        "ROLE": "acl_role",
                        "SET": "resource_set",
                    },
                    removing_references_from={
                        "A": {"T", "SET"},
                        "PERMISSION": {"ROLE"},
                    },
                ),
            ]
        )

    def test_remove_resources_started(self):
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="P-1"/>
                    <primitive id="P-2"/>
                </resources>
            """
        )
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <resource id="P-1" role="Started"/>
                    <resource id="P-2" role="Stopped"/>
                </resources>
            """
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_elements(
                self.env_assist.get_env(), ["P-1", "P-2"]
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.CANNOT_REMOVE_RESOURCES_NOT_STOPPED,
                    resource_id_list=["P-1"],
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_remove_resources_started_forced(self):
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="P-1"/>
                    <primitive id="P-2"/>
                </resources>
            """
        )
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <resource id="P-1" role="Started"/>
                    <resource id="P-2" role="Stopped"/>
                </resources>
            """
        )
        self.config.env.push_cib(resources="<resources />")

        lib.remove_elements(
            self.env_assist.get_env(), ["P-1", "P-2"], [reports.codes.FORCE]
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.CANNOT_REMOVE_RESOURCES_NOT_STOPPED,
                    resource_id_list=["P-1"],
                )
            ]
        )

    def test_remove_resources_stonith_state_check_skipped(self):
        # adding extra stonith, so the last stonith check passes
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="P-1"/>
                    <primitive class="stonith" id="S-1" type="fence_xvm"/>
                    <primitive class="stonith" id="S-2" type="fence_xvm"/>
                </resources>
            """
        )
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <resource id="P-1" role="Stopped"/>
                    <resource id="S-1" role="Started"/>
                </resources>
            """
        )
        self.config.env.push_cib(
            resources="""
                <resources>
                    <primitive class="stonith" id="S-2" type="fence_xvm"/>
                </resources>
            """
        )

        lib.remove_elements(self.env_assist.get_env(), ["P-1", "S-1"])

    def test_remove_resources_not_live_cib(self):
        cib = modify_cib(
            read_test_resource("cib-empty.xml"),
            resources="""
                <resources>
                    <primitive id="A"/>
                </resources>
            """,
        )
        # This makes env.is_cib_live return False
        self.config.env.set_cib_data(cib)
        self.config.runner.cib.load_content(
            cib, env={"CIB_file": "/fake/tmp/file"}
        )
        self.config.env.push_cib(
            resources="<resources/>", load_key="runner.cib.load_content"
        )

        lib.remove_elements(self.env_assist.get_env(), ["A"])
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.STOPPED_RESOURCES_BEFORE_DELETE_CHECK_SKIPPED,
                    resource_id_list=["A"],
                    reason_type=reports.const.REASON_NOT_LIVE_CIB,
                )
            ]
        )


class StonithAndSbdCheck(TestCase):
    """
    Test that an error is produced when removing the last stonith resource and
    sbd is disabled
    """

    resources = """
        <resources>
            <primitive id="S1" class="stonith" type="fence_any" />
            <primitive id="S2" class="stonith" type="fence_any" />
        </resources>
    """

    def fixture_config_sbd_calls(self, sbd_enabled):
        node_name_list = ["node-1", "node-2"]
        self.config.env.set_known_nodes(node_name_list)
        self.config.corosync_conf.load(node_name_list=node_name_list)
        self.config.http.sbd.check_sbd(
            communication_list=[
                dict(
                    label=node,
                    param_list=[("watchdog", ""), ("device_list", "[]")],
                    output=json.dumps(
                        dict(
                            sbd=dict(
                                installed=True,
                                enabled=sbd_enabled,
                                running=sbd_enabled,
                            )
                        )
                    ),
                )
                for node in node_name_list
            ]
        )

    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_some_stonith_left(self):
        # sbd calls do not happen if some stonith is left
        self.config.runner.cib.load(resources=self.resources)
        self.config.env.push_cib(
            resources="""
                <resources>
                    <primitive id="S2" class="stonith" type="fence_any" />
                </resources>
            """
        )

        lib.remove_elements(self.env_assist.get_env(), ["S1"])
        self.env_assist.assert_reports([])

    def test_no_stonith_left_sbd_enabled(self):
        self.config.runner.cib.load(resources=self.resources)
        self.fixture_config_sbd_calls(sbd_enabled=True)
        self.config.env.push_cib(resources="<resources />")

        lib.remove_elements(self.env_assist.get_env(), ["S1", "S2"])
        self.env_assist.assert_reports([])

    def test_fake_stonith_left_sbd_disabled(self):
        resources = """
            <resources>
                <primitive id="S1" class="stonith" type="fence_any" />
                <primitive id="S2" class="stonith" type="fence_sbd" />
            </resources>
        """
        self.config.runner.cib.load(resources=resources)
        self.fixture_config_sbd_calls(sbd_enabled=False)

        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_elements(self.env_assist.get_env(), ["S1"])
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.NO_STONITH_MEANS_WOULD_BE_LEFT,
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_disabled_stonith_left_sbd_disabled(self):
        resources = """
            <resources>
                <primitive id="S1" class="stonith" type="fence_any">
                    <meta_attributes>
                        <nvpair name="target-role" value="stopped" />
                    </meta_attributes>
                </primitive>
                <primitive id="S2" class="stonith" type="fence_any" />
            </resources>
        """
        self.config.runner.cib.load(resources=resources)
        self.fixture_config_sbd_calls(sbd_enabled=False)

        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_elements(self.env_assist.get_env(), ["S2"])
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.NO_STONITH_MEANS_WOULD_BE_LEFT,
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_stonith_was_already_noneffective_and_sbd_disabled(self):
        resources = """
            <resources>
                <primitive id="S1" class="stonith" type="fence_any">
                    <meta_attributes id="S1-meta_attributes">
                        <nvpair id="S1-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                </primitive>
                <primitive id="S2" class="stonith" type="fence_sbd" />
            </resources>
        """
        self.config.runner.cib.load(resources=resources)
        self.fixture_config_sbd_calls(sbd_enabled=False)
        self.config.env.push_cib(resources="<resources />")

        lib.remove_elements(self.env_assist.get_env(), ["S1", "S2"])
        self.env_assist.assert_reports([])

    def test_no_stonith_left_sbd_disabled(self):
        self.config.runner.cib.load(resources=self.resources)
        self.fixture_config_sbd_calls(sbd_enabled=False)

        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_elements(self.env_assist.get_env(), ["S1", "S2"])
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.NO_STONITH_MEANS_WOULD_BE_LEFT,
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_no_stonith_left_sbd_disabled_forced(self):
        self.config.runner.cib.load(resources=self.resources)
        self.fixture_config_sbd_calls(sbd_enabled=False)
        self.config.env.push_cib(resources="<resources />")

        lib.remove_elements(
            self.env_assist.get_env(), ["S1", "S2"], {reports.codes.FORCE}
        )

        self.env_assist.assert_reports(
            [fixture.warn(reports.codes.NO_STONITH_MEANS_WOULD_BE_LEFT)]
        )

    def test_no_stonith_left_sbd_partially_enabled(self):
        node_name_list = ["node-1", "node-2"]
        self.config.env.set_known_nodes(node_name_list)
        self.config.runner.cib.load(resources=self.resources)
        self.config.corosync_conf.load(node_name_list=node_name_list)
        self.config.http.sbd.check_sbd(
            communication_list=[
                dict(
                    label="node-1",
                    param_list=[("watchdog", ""), ("device_list", "[]")],
                    output=json.dumps(
                        dict(
                            sbd=dict(installed=True, enabled=True, running=True)
                        )
                    ),
                ),
                dict(
                    label="node-2",
                    param_list=[("watchdog", ""), ("device_list", "[]")],
                    output=json.dumps(
                        dict(
                            sbd=dict(
                                installed=True, enabled=False, running=False
                            )
                        )
                    ),
                ),
            ]
        )
        self.config.env.push_cib(resources="<resources />")

        lib.remove_elements(self.env_assist.get_env(), ["S1", "S2"])
        self.env_assist.assert_reports([])

    def test_communication_error(self):
        self.config.runner.cib.load(resources=self.resources)
        node_name_list = ["node-1", "node-2"]
        self.config.env.set_known_nodes(node_name_list)
        self.config.corosync_conf.load(node_name_list=node_name_list)
        self.config.http.sbd.check_sbd(
            communication_list=[
                dict(
                    label="node-1",
                    param_list=[("watchdog", ""), ("device_list", "[]")],
                    output=json.dumps(
                        dict(
                            sbd=dict(installed=True, enabled=True, running=True)
                        )
                    ),
                ),
                dict(
                    label="node-2",
                    param_list=[("watchdog", ""), ("device_list", "[]")],
                    was_connected=False,
                ),
            ]
        )
        self.config.env.push_cib(resources="<resources />")

        lib.remove_elements(self.env_assist.get_env(), ["S1", "S2"])
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node="node-2",
                    command="remote/check_sbd",
                    reason=None,
                ),
                fixture.warn(
                    reports.codes.UNABLE_TO_GET_SBD_STATUS,
                    node="node-2",
                    reason="",
                ),
            ]
        )


class ElementDescriptionSet(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.new_description = "I am a stick."
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <primitive id="B" description="I am a description"/>
                </resources>
            """
        )

    def test_success_add_description(self):
        self.config.env.push_cib(
            resources=f"""
                <resources>
                    <primitive id="A" description="{self.new_description}"/>
                    <primitive id="B" description="I am a description"/>
                </resources>
            """
        )
        lib.element_description_set(
            self.env_assist.get_env(), "A", self.new_description
        )

    def test_success_update_description(self):
        self.config.env.push_cib(
            resources=f"""
                <resources>
                    <primitive id="A"/>
                    <primitive id="B" description="{self.new_description}"/>
                </resources>
            """
        )
        lib.element_description_set(
            self.env_assist.get_env(), "B", self.new_description
        )

    def test_success_remove_description(self):
        self.config.env.push_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <primitive id="B"/>
                </resources>
            """
        )
        lib.element_description_set(self.env_assist.get_env(), "B", "")

    def test_element_does_not_exist(self):
        self.env_assist.assert_raise_library_error(
            lambda: lib.element_description_set(
                self.env_assist.get_env(), "C", self.new_description
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="C",
                    expected_types=[],
                    context_type="",
                    context_id="",
                )
            ]
        )

    def test_element_does_not_support_description(self):
        self.config.runner.cib.load(
            instead="runner.cib.load",
            constraints="""
                <constraints>
                    <rsc_location id="L1" rsc="A" node="node1" score="200"/>
                </constraints>
            """,
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.element_description_set(
                self.env_assist.get_env(), "L1", self.new_description
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_DOES_NOT_SUPPORT_ELEMENT_DESCRIPTIONS,
                    element_id="L1",
                    element_type="rsc_location",
                    expected_types=sorted(TAG_LIST_SUPPORTS_DESCRIPTION),
                ),
            ]
        )


class ElementDescriptionGet(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.description = "I am a stick."
        self.config.runner.cib.load(
            resources=f"""
                <resources>
                    <primitive id="A" description="{self.description}"/>
                    <primitive id="B"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_location id="L1" rsc="A" node="node1" score="200"/>
                </constraints>
            """,
        )

    def test_success_return_description(self):
        result = lib.element_description_get(self.env_assist.get_env(), "A")
        self.assertEqual(result, self.description)

    def test_success_no_description(self):
        result = lib.element_description_get(self.env_assist.get_env(), "B")
        self.assertEqual(result, "")

    def test_element_does_not_exist(self):
        self.env_assist.assert_raise_library_error(
            lambda: lib.element_description_get(self.env_assist.get_env(), "C")
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="C",
                    expected_types=[],
                    context_type="",
                    context_id="",
                )
            ]
        )

    def test_element_does_not_support_description(self):
        self.env_assist.assert_raise_library_error(
            lambda: lib.element_description_get(self.env_assist.get_env(), "L1")
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_DOES_NOT_SUPPORT_ELEMENT_DESCRIPTIONS,
                    element_id="L1",
                    element_type="rsc_location",
                    expected_types=sorted(TAG_LIST_SUPPORTS_DESCRIPTION),
                ),
            ]
        )



from unittest import TestCase, mock

from pcs.cli.reports.output import process_library_reports
from pcs.common import reports


@mock.patch("pcs.cli.reports.output.print_to_stderr")
class ProcessLibraryReports(TestCase):
    report_message = reports.messages.NoActionNecessary()
    report_message_text = report_message.message

    def test_no_reports(self, mock_print_to_stderr):
        with self.assertRaises(SystemExit) as cm:
            process_library_reports([])
        self.assertEqual(cm.exception.code, 1)
        mock_print_to_stderr.assert_called_once_with(
            "Error: Errors have occurred, therefore pcs is unable to continue"
        )

    def test_all_report_severities_no_exit_on_error(self, mock_print_to_stderr):
        report_items = [
            reports.ReportItem.debug(self.report_message),
            reports.ReportItem.info(self.report_message),
            reports.ReportItem.warning(self.report_message),
            reports.ReportItem.deprecation(self.report_message),
            reports.ReportItem.error(self.report_message),
        ]

        process_library_reports(report_items, exit_on_error=False)

        mock_print_to_stderr.assert_has_calls(
            [
                mock.call(self.report_message_text),
                mock.call(self.report_message_text),
                mock.call(f"Warning: {self.report_message_text}"),
                mock.call(f"Deprecation Warning: {self.report_message_text}"),
                mock.call(f"Error: {self.report_message_text}"),
            ]
        )
        self.assertEqual(mock_print_to_stderr.call_count, 5)

    def test_exit_on_error(self, mock_print_to_stderr):
        with self.assertRaises(SystemExit) as cm:
            process_library_reports(
                [reports.ReportItem.error(self.report_message)],
                exit_on_error=True,
            )
        self.assertEqual(cm.exception.code, 1)
        mock_print_to_stderr.assert_called_once_with(
            f"Error: {self.report_message_text}"
        )

    def test_error_with_force_code(self, mock_print_to_stderr):
        with self.assertRaises(SystemExit) as cm:
            process_library_reports(
                [
                    reports.ReportItem.error(
                        self.report_message, force_code=reports.codes.FORCE
                    )
                ],
                exit_on_error=True,
            )
        self.assertEqual(cm.exception.code, 1)
        mock_print_to_stderr.assert_called_once_with(
            f"Error: {self.report_message_text}, use --force to override"
        )

    def test_context(self, mock_print_to_stderr):
        report_items = [
            reports.ReportItem.info(
                self.report_message, context=reports.ReportItemContext("node-A")
            ),
            reports.ReportItem.error(
                self.report_message, context=reports.ReportItemContext("node-B")
            ),
        ]

        process_library_reports(report_items, exit_on_error=False)

        mock_print_to_stderr.assert_has_calls(
            [
                mock.call(f"node-A: {self.report_message_text}"),
                mock.call(f"Error: node-B: {self.report_message_text}"),
            ]
        )
        self.assertEqual(mock_print_to_stderr.call_count, 2)

    def test_do_not_include_debug(self, mock_print_to_stderr):
        report_items = [
            reports.ReportItem.debug(self.report_message),
            reports.ReportItem.warning(self.report_message),
        ]

        process_library_reports(report_items, include_debug=False)

        mock_print_to_stderr.assert_called_once_with(
            f"Warning: {self.report_message_text}"
        )



import abc
from typing import Iterable, Mapping, Sequence, TypedDict

from lxml.etree import SubElement, _Element

from pcs.common import reports
from pcs.common.types import StringSequence
from pcs.lib.cib.const import (
    TAG_LIST_CONSTRAINABLE,
    TAG_LIST_CONSTRAINT,
    TAG_LIST_RESOURCE_MULTIINSTANCE,
    TAG_RESOURCE_SET,
)
from pcs.lib.cib.constraint import resource_set
from pcs.lib.cib.tools import IdProvider, Version
from pcs.lib.pacemaker.values import sanitize_id
from pcs.lib.validate import TypeOptionMap
from pcs.lib.xml_tools import find_parent


# structure of a parameter of a command for creating set constraints
class CmdInputResourceSet(TypedDict):
    ids: StringSequence
    options: Mapping[str, str]


# structure of a parameter of a command for creating set constraints
CmdInputResourceSetList = Sequence[CmdInputResourceSet]


# structure of a parameter of a command for creating set constraints prepared
# for further validation and processing
class CmdInputResourceSetLoaded(TypedDict):
    constrained_elements: Sequence[_Element]
    options: TypeOptionMap


# structure of a parameter of a command for creating set constraints prepared
# for further validation and processing
CmdInputResourceSetLoadedList = list[CmdInputResourceSetLoaded]


def is_constraint(element: _Element) -> bool:
    return element.tag in TAG_LIST_CONSTRAINT


def is_set_constraint(element: _Element) -> bool:
    return (
        is_constraint(element)
        and element.find(f"./{TAG_RESOURCE_SET}") is not None
    )


def find_constraints_of_same_type(
    constraint_section: _Element,
    constraint_to_check: _Element,
) -> Iterable[_Element]:
    """
    Find constraints of the same type and setness as a specified constraint

    constraint_section -- where to look for constraints
    constraint_to_check -- defines a type and a setness to look for
    """
    looking_for_set_constraint = is_set_constraint(constraint_to_check)
    return (
        element
        for element in constraint_section.iterfind(constraint_to_check.tag)
        if element is not constraint_to_check
        and is_set_constraint(element) == looking_for_set_constraint
    )


def _create_set_constraint_id(
    id_provider: IdProvider, prefix: str, rsc_set_list: CmdInputResourceSetList
) -> str:
    # Create a semi-random id. We need it to be predictable (for testing), short
    # and somehow different than other ids so that we don't spend much time in
    # finding a unique id
    # Avoid using actual resource names. It makes the id very long (consider 10
    # or more resources in a set constraint). Also, if a resource is deleted
    # and therefore removed from the constraint, the id no longer matches the
    # constraint.
    resource_ids: list[str] = []
    for rsc_set in rsc_set_list:
        resource_ids += rsc_set["ids"]
    id_part = "".join([_id[0] + _id[-1] for _id in resource_ids][:3])
    return id_provider.allocate_id(sanitize_id(f"{prefix}_set_{id_part}"))


def create_constraint_with_set(
    parent_element: _Element,
    id_provider: IdProvider,
    cib_schema_version: Version,
    constraint_tag: str,
    id_prefix: str,
    rsc_set_list: CmdInputResourceSetList,
    options: Mapping[str, str],
) -> _Element:
    """
    Create a constraint with resource sets

    parent_element -- where to place the constraint
    id_provider -- elements' ids generator
    cib_schema_version -- current CIB schema version
    constraint_tag -- constraint xml tag
    id_prefix -- constraint id prefix
    rsc_set_list -- definition of resource sets: resources and set options
    options -- additional options for the constraint
    """
    constraint_el = SubElement(parent_element, constraint_tag)

    if "id" not in options:
        constraint_el.attrib["id"] = _create_set_constraint_id(
            id_provider, id_prefix, rsc_set_list
        )

    for name, value in options.items():
        if value != "":
            constraint_el.attrib[name] = value

    for rsc_set in rsc_set_list:
        resource_set.create(
            constraint_el,
            id_provider,
            cib_schema_version,
            rsc_set["ids"],
            rsc_set["options"],
        )

    return constraint_el


class DuplicatesChecker:
    """
    Base class for finding duplicate constraints

    To use it, create a subclass and implement _are_duplicate method to compare
    constraints of a specific type
    """

    def __init__(self) -> None:
        pass

    def check(
        self,
        constraint_section: _Element,
        constraint_to_check: _Element,
        force_flags: reports.types.ForceFlags = (),
    ) -> reports.ReportItemList:
        """
        Check if a constraint is a duplicate of an already existing constraint

        constraint_section -- where to look for existing constraints
        constraint_to_check -- search for duplicates of this constraint
        force_flags -- list of flags codes
        """
        self._check_init(constraint_to_check)
        report_list: reports.ReportItemList = []
        duplication_allowed = reports.codes.FORCE in force_flags

        duplicate_constraint_list = [
            constraint_el
            for constraint_el in find_constraints_of_same_type(
                constraint_section, constraint_to_check
            )
            if self._are_duplicate(constraint_to_check, constraint_el)
        ]

        if duplicate_constraint_list:
            report_list.append(
                reports.ReportItem(
                    severity=reports.item.get_severity(
                        reports.codes.FORCE, duplication_allowed
                    ),
                    message=reports.messages.DuplicateConstraintsExist(
                        [
                            str(duplicate_el.attrib["id"])
                            for duplicate_el in duplicate_constraint_list
                        ]
                    ),
                ),
            )

        return report_list

    def _check_init(self, constraint_to_check: _Element) -> None:
        """
        For descendants to do their initialization for each check
        """

    @abc.abstractmethod
    def _are_duplicate(
        self,
        constraint_to_check: _Element,
        constraint_el: _Element,
    ) -> bool:
        """
        Compare two constraints and decide if they are duplicate to each other

        constraint_to_check -- search for duplicates of this constraint
        constraint_el -- an already existing constraint
        """
        raise NotImplementedError()


class DuplicatesCheckerSetConstraint(DuplicatesChecker):
    _constraint_id_set_list: list[list[str]]

    @staticmethod
    def _get_id_set_list(constraint_el: _Element) -> list[list[str]]:
        return [
            resource_set.get_resource_id_set_list(resource_set_item)
            for resource_set_item in constraint_el.findall(
                f".//{resource_set.TAG_RESOURCE_SET}"
            )
        ]

    def _check_init(self, constraint_to_check: _Element) -> None:
        self._constraint_id_set_list = self._get_id_set_list(
            constraint_to_check
        )

    def _are_duplicate(
        self,
        constraint_to_check: _Element,
        constraint_el: _Element,
    ) -> bool:
        del constraint_to_check
        return (
            self._get_id_set_list(constraint_el) == self._constraint_id_set_list
        )


def validate_constrainable_elements(
    element_list: Iterable[_Element], in_multiinstance_allowed: bool = False
) -> reports.ReportItemList:
    """
    Validate that a constraint can be created for each of the specified elements

    element_list -- the elements to be validated
    in_multiinstance_allowed -- allow constraints for resources in clones/bundles
    """
    report_list: reports.ReportItemList = []

    for element in element_list:
        if element.tag not in TAG_LIST_CONSTRAINABLE:
            report_list.append(
                reports.ReportItem.error(
                    reports.messages.IdBelongsToUnexpectedType(
                        str(element.attrib["id"]),
                        sorted(TAG_LIST_CONSTRAINABLE),
                        element.tag,
                    )
                )
            )
            continue

        if element.tag not in TAG_LIST_RESOURCE_MULTIINSTANCE:
            multiinstance_parent = find_parent(
                element, TAG_LIST_RESOURCE_MULTIINSTANCE
            )
            if multiinstance_parent is not None:
                report_list.append(
                    reports.ReportItem(
                        reports.item.get_severity(
                            reports.codes.FORCE, in_multiinstance_allowed
                        ),
                        reports.messages.ResourceForConstraintIsMultiinstance(
                            str(element.attrib["id"]),
                            multiinstance_parent.tag,
                            str(multiinstance_parent.attrib["id"]),
                        ),
                    )
                )

    return report_list



import os.path
import sys
from textwrap import dedent

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(CURRENT_DIR, "{}.d".format(sys.argv[0]))


def get_arg_values(argv, name):
    values = []
    next_value = len(argv)
    for i, value in enumerate(argv):
        if value == name:
            next_value = i + 1
        elif i == next_value:
            values.append(value)
    return values


def main():
    argv = sys.argv[1:]
    if not argv:
        raise AssertionError()

    arg = argv.pop(0)

    if arg == "--validate":
        if get_arg_values(argv, "--output-as")[0] != "xml":
            raise AssertionError()
        is_invalid = False
        for arg in argv:
            if "=" in arg and arg.split("=", 1)[1] == "is_invalid=True":
                is_invalid = True
                break
        output = ""
        if is_invalid:
            output = """<output source="stderr">pcsmock validation failure</output>"""
        cmd_str = " ".join(sys.argv)
        agent_type = get_arg_values(argv, "--agent")[0]
        stdout = dedent(
            f"""
            <pacemaker-result api-version="2.15" request="{cmd_str}">
              <validate agent="{agent_type}" valid="{str(is_invalid).lower()}">
                <command code="{-201 if is_invalid else 0}">
                {output}
                </command>
              </validate>
              <status code="{1 if is_invalid else 0}" message="{"Error" if is_invalid else "OK"}"/>
            </pacemaker-result>
            """
        )
        sys.stdout.write(stdout)
        if is_invalid:
            raise SystemExit(1)
    else:
        raise AssertionError()


if __name__ == "__main__":
    main()



from unittest import TestCase, mock

from pcs import settings
from pcs.lib.pcs_cfgsync.config.facade import Facade as CfgsyncFacade


class Facade(TestCase):
    def assert_all_default_values(self, facade: CfgsyncFacade):
        self.assertTrue(facade.is_sync_allowed)
        self.assertFalse(facade.is_sync_paused)
        self.assertEqual(
            settings.pcs_cfgsync_thread_interval_default, facade.sync_interval
        )
        self.assertEqual(
            settings.pcs_cfgsync_thread_interval_previous_not_connected_default,
            facade.sync_interval_previous_not_connected,
        )
        self.assertEqual(
            settings.pcs_cfgsync_file_backup_count_default,
            facade.file_backup_count,
        )

    def test_create(self):
        facade = CfgsyncFacade.create()
        self.assert_all_default_values(facade)

    def test_empty_file(self):
        facade = CfgsyncFacade({})
        self.assert_all_default_values(facade)

    def test_file_with_values(self):
        facade = CfgsyncFacade(
            {
                "thread_disabled": True,
                "thread_interval": 100,
                "thread_interval_previous_not_connected": 30,
                "file_backup_count": 20,
            }
        )
        self.assertFalse(facade.is_sync_allowed)
        self.assertFalse(facade.is_sync_paused)
        self.assertEqual(100, facade.sync_interval)
        self.assertEqual(30, facade.sync_interval_previous_not_connected)
        self.assertEqual(20, facade.file_backup_count)

    def test_too_low_values(self):
        facade = CfgsyncFacade(
            {
                "thread_interval": 1,
                "thread_interval_previous_not_connected": 1,
                "file_backup_count": -1,
            }
        )
        self.assertEqual(
            settings.pcs_cfgsync_thread_interval_minimum, facade.sync_interval
        )
        self.assertEqual(
            settings.pcs_cfgsync_thread_interval_previous_not_connected_minimum,
            facade.sync_interval_previous_not_connected,
        )
        self.assertEqual(
            settings.pcs_cfgsync_file_backup_count_minimum,
            facade.file_backup_count,
        )

    @mock.patch("pcs.lib.pcs_cfgsync.config.facade.time.time", lambda: 1000)
    def test_paused_still_paused(self):
        facade = CfgsyncFacade({"thread_paused_until": 2000})
        self.assertFalse(facade.is_sync_allowed)
        self.assertTrue(facade.is_sync_paused)

    @mock.patch("pcs.lib.pcs_cfgsync.config.facade.time.time", lambda: 1000)
    def test_paused_no_longer_paused(self):
        facade = CfgsyncFacade({"thread_paused_until": 500})
        self.assertTrue(facade.is_sync_allowed)
        self.assertFalse(facade.is_sync_paused)

    @mock.patch("pcs.lib.pcs_cfgsync.config.facade.time.time", lambda: 1000)
    def test_file_with_incorrect_values(self):
        facade = CfgsyncFacade(
            {
                "thread_disabled": "a",
                "thread_interval": "b",
                "thread_interval_previous_not_connected": "c",
                "file_backup_count": "d",
                "thread_paused_until": "e",
            }
        )
        self.assert_all_default_values(facade)



from unittest import TestCase, mock

from pcs.common import reports
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
    ListCibNvsetDto,
)
from pcs.lib.commands import cluster_property

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from .fixtures import fixture_crm_config_properties


class TestGetProperties(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def command(self, evaluate_expired=False):
        return cluster_property.get_properties(
            self.env_assist.get_env(),
            evaluate_expired=evaluate_expired,
        )

    def test_no_properties_configured(self):
        self.config.runner.cib.load()
        self.assertEqual(self.command(), ListCibNvsetDto(nvsets=[]))

    def test_empty_cluster_property_set(self):
        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties([("set_id", {})])
        )
        self.assertEqual(
            self.command(),
            ListCibNvsetDto(
                nvsets=[
                    CibNvsetDto(id="set_id", options={}, rule=None, nvpairs=[])
                ]
            ),
        )

    def test_cluster_property_set_with_properties(self):
        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties(
                [("set_id", {"prop1": "val1", "prop2": "val2"})],
                score_list=[100],
            )
        )
        self.assertEqual(
            self.command(),
            ListCibNvsetDto(
                nvsets=[
                    CibNvsetDto(
                        id="set_id",
                        options={"score": "100"},
                        rule=None,
                        nvpairs=[
                            CibNvpairDto(
                                id="set_id-prop1", name="prop1", value="val1"
                            ),
                            CibNvpairDto(
                                id="set_id-prop2", name="prop2", value="val2"
                            ),
                        ],
                    )
                ]
            ),
        )

    def test_more_cluster_property_sets(self):
        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties(
                [
                    ("set_id", {"prop1": "val1", "prop2": "val2"}),
                    ("set_id2", {"prop3": "val3", "prop4": "val4"}),
                ],
                score_list=[100, 200],
            )
        )
        self.assertEqual(
            self.command(),
            ListCibNvsetDto(
                nvsets=[
                    CibNvsetDto(
                        id="set_id",
                        options={"score": "100"},
                        rule=None,
                        nvpairs=[
                            CibNvpairDto(
                                id="set_id-prop1", name="prop1", value="val1"
                            ),
                            CibNvpairDto(
                                id="set_id-prop2", name="prop2", value="val2"
                            ),
                        ],
                    ),
                    CibNvsetDto(
                        id="set_id2",
                        options={"score": "200"},
                        rule=None,
                        nvpairs=[
                            CibNvpairDto(
                                id="set_id2-prop3", name="prop3", value="val3"
                            ),
                            CibNvpairDto(
                                id="set_id2-prop4", name="prop4", value="val4"
                            ),
                        ],
                    ),
                ]
            ),
        )

    def test_cib_error(self):
        self.config.runner.cib.load(returncode=1, stderr="error")
        self.env_assist.assert_raise_library_error(
            self.command,
            reports=[
                fixture.error(reports.codes.CIB_LOAD_ERROR, reason="error")
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports([])

    @mock.patch(
        "pcs.lib.cib.rule.in_effect.has_rule_in_effect_status_tool",
        lambda: True,
    )
    def test_evaluate_expired_but_no_set_rule(self):
        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties([("set_id", {})])
        )
        self.assertEqual(
            self.command(evaluate_expired=True),
            ListCibNvsetDto(
                nvsets=[
                    CibNvsetDto(id="set_id", options={}, rule=None, nvpairs=[])
                ]
            ),
        )

    @mock.patch(
        "pcs.lib.cib.rule.in_effect.has_rule_in_effect_status_tool",
        lambda: False,
    )
    def test_evaluate_expired_no_status_tool(self):
        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties([("set_id", {})])
        )
        self.assertEqual(
            self.command(evaluate_expired=True),
            ListCibNvsetDto(
                nvsets=[
                    CibNvsetDto(id="set_id", options={}, rule=None, nvpairs=[])
                ]
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RULE_IN_EFFECT_STATUS_DETECTION_NOT_SUPPORTED,
                )
            ]
        )



from typing import Mapping

from lxml.etree import SubElement, _Element

from pcs.common import const, pacemaker, reports
from pcs.common.const import PcmkAction
from pcs.common.pacemaker.constraint import CibResourceSetDto
from pcs.common.pacemaker.role import (
    get_value_for_cib as get_role_value_for_cib,
)
from pcs.common.pacemaker.types import (
    CibResourceSetOrdering,
    CibResourceSetOrderType,
)
from pcs.common.types import StringIterable
from pcs.lib import validate
from pcs.lib.cib.const import TAG_RESOURCE_REF, TAG_RESOURCE_SET
from pcs.lib.cib.resource import group
from pcs.lib.cib.resource.common import get_parent_resource
from pcs.lib.cib.tools import (
    IdProvider,
    Version,
    are_new_role_names_supported,
    find_unique_id,
    get_elements_by_ids,
    role_constructor,
)
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import is_true
from pcs.lib.tools import get_optional_value

_ATTRIBUTES = ("action", "require-all", "role", "sequential")


# DEPRECATED
def prepare_set(
    find_valid_id, resource_set, report_processor: reports.ReportProcessor
):
    """return resource_set with corrected ids"""
    if report_processor.report_list(
        _validate_options(resource_set["options"])
    ).has_errors:
        raise LibraryError()
    return {
        "ids": [find_valid_id(id_) for id_ in resource_set["ids"]],
        "options": resource_set["options"],
    }


# DEPRECATED pacemaker cares, thus the validation should be constraint specific
def _validate_options(options) -> reports.ReportItemList:
    # Pacemaker does not care currently about meaningfulness for concrete
    # constraint, so we use all attribs.
    validators = [
        validate.NamesIn(_ATTRIBUTES, option_type="set"),
        validate.ValueIn("action", const.PCMK_ACTIONS),
        validate.ValuePcmkBoolean("require-all"),
        validate.ValueIn("role", const.PCMK_ROLES),
        validate.ValuePcmkBoolean("sequential"),
    ]
    return validate.ValidatorAll(validators).validate(options)


def create(
    parent_el: _Element,
    id_provider: IdProvider,
    cib_schema_version: Version,
    rsc_list: StringIterable,
    set_options: Mapping[str, str],
) -> _Element:
    """
    Create a new resource set element

    parent_el -- where to create the set
    id_provider -- elements' ids generator
    cib_schema_version -- current CIB schema version
    rsc_list -- list of resources in the set
    set_options -- set attributes
    """
    rsc_set_el = SubElement(parent_el, TAG_RESOURCE_SET)

    if "id" not in set_options:
        rsc_set_el.attrib["id"] = id_provider.allocate_id(
            "{0}_set".format(parent_el.attrib.get("id", "constraint_set")),
        )

    for name, value in set_options.items():
        if name == "role":
            # noqa - for loop variable 'value' overwritten by assignment target
            value = get_role_value_for_cib(  # noqa: PLW2901
                const.PcmkRoleType(value),
                cib_schema_version >= const.PCMK_NEW_ROLES_CIB_VERSION,
            )
        if value != "":
            rsc_set_el.attrib[name] = value

    for rsc_id in rsc_list:
        SubElement(rsc_set_el, TAG_RESOURCE_REF).attrib["id"] = rsc_id

    return rsc_set_el


# DEPRECATED
def create_old(parent, resource_set):
    """
    parent - lxml element for append new resource_set
    """
    element = SubElement(parent, "resource_set")
    if "role" in resource_set["options"]:
        resource_set["options"]["role"] = pacemaker.role.get_value_for_cib(
            resource_set["options"]["role"],
            is_latest_supported=are_new_role_names_supported(parent),
        )
    element.attrib.update(resource_set["options"])
    element.attrib["id"] = find_unique_id(
        parent.getroottree(),
        "{0}_set".format(parent.attrib.get("id", "constraint_set")),
    )

    for _id in resource_set["ids"]:
        SubElement(element, "resource_ref").attrib["id"] = _id

    return element


def get_resource_id_set_list(element: _Element) -> list[str]:
    return [
        str(resource_ref_element.attrib["id"])
        for resource_ref_element in element.findall(f".//{TAG_RESOURCE_REF}")
    ]


def is_resource_in_same_group(cib, resource_id_list):
    # We don't care about not found elements here, that is a job of another
    # validator. We do not care if the id doesn't belong to a resource either
    # for the same reason.
    element_list, _ = get_elements_by_ids(cib, set(resource_id_list))

    parent_list = []
    for element in element_list:
        parent = get_parent_resource(element)
        if parent is not None and group.is_group(parent):
            parent_list.append(parent)

    if len(set(parent_list)) != len(parent_list):
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.CannotSetOrderConstraintsForResourcesInTheSameGroup()
            )
        )


def _resource_set_element_to_dto(
    resource_set_el: _Element,
) -> CibResourceSetDto:
    return CibResourceSetDto(
        set_id=resource_set_el.get("id", ""),
        sequential=get_optional_value(
            is_true, resource_set_el.get("sequential")
        ),
        require_all=get_optional_value(
            is_true, resource_set_el.get("require-all")
        ),
        ordering=get_optional_value(
            CibResourceSetOrdering, resource_set_el.get("ordering")
        ),
        action=get_optional_value(PcmkAction, resource_set_el.get("action")),
        role=get_optional_value(role_constructor, resource_set_el.get("role")),
        score=resource_set_el.get("score"),
        kind=get_optional_value(
            CibResourceSetOrderType, resource_set_el.get("kind")
        ),
        resources_ids=[
            str(rsc_ref.attrib["id"])
            for rsc_ref in resource_set_el.findall(f"./{TAG_RESOURCE_REF}")
        ],
    )


def constraint_element_to_resource_set_dto_list(
    constraint_el: _Element,
) -> list[CibResourceSetDto]:
    return [
        _resource_set_element_to_dto(set_el)
        for set_el in constraint_el.findall(f"./{TAG_RESOURCE_SET}")
    ]
